"""Test client sync."""
from itertools import count
import ssl

from test.conftest import mockSocket
import unittest
from unittest.mock import MagicMock, Mock, patch

import serial

from pymodbus.client import (
    ModbusSerialClient,
    ModbusTcpClient,
    ModbusTlsClient,
    ModbusUdpClient,
)
from pymodbus.client.tls import sslctx_provider
from pymodbus.exceptions import ConnectionException
from pymodbus.transaction import (
    ModbusAsciiFramer,
    ModbusBinaryFramer,
    ModbusRtuFramer,
    ModbusSocketFramer,
    ModbusTlsFramer,
)


# ---------------------------------------------------------------------------#
# Fixture
# ---------------------------------------------------------------------------#


class SynchronousClientTest(
    unittest.TestCase
):  # pylint: disable=too-many-public-methods
    """Unittest for the pymodbus.client module."""

    # -----------------------------------------------------------------------#
    # Test UDP Client
    # -----------------------------------------------------------------------#

    def test_basic_syn_udp_client(self):
        """Test the basic methods for the udp sync client"""
        # receive/send
        client = ModbusUdpClient("127.0.0.1")
        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(1, client.send(b"\x50"))
        self.assertEqual(b"\x50", client.recv(1))

        # connect/disconnect
        self.assertTrue(client.connect())
        client.close()

        # already closed socket
        client.socket = False
        client.close()

        self.assertEqual("ModbusUdpClient(127.0.0.1:502)", str(client))

    def test_udp_client_is_socket_open(self):
        """Test the udp client is_socket_open method"""
        client = ModbusUdpClient("127.0.0.1")
        self.assertTrue(client.is_socket_open())

    def test_udp_client_send(self):
        """Test the udp client send method"""
        client = ModbusUdpClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.send(None),
        )

        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(4, client.send("1234"))

    def test_udp_client_recv(self):
        """Test the udp client receive method"""
        client = ModbusUdpClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )

        client.socket = mockSocket()
        client.socket.mock_store(b"\x00" * 4)
        self.assertEqual(b"", client.recv(0))
        self.assertEqual(b"\x00" * 4, client.recv(4))

    def test_udp_client_repr(self):
        """Test udp client representation."""
        client = ModbusUdpClient("127.0.0.1")
        rep = (
            f"<{client.__class__.__name__} at {hex(id(client))} socket={client.socket}, "
            f"ipaddr={client.params.host}, port={client.params.port}, timeout={client.params.timeout}>"
        )
        self.assertEqual(repr(client), rep)

    # -----------------------------------------------------------------------#
    # Test TCP Client
    # -----------------------------------------------------------------------#

    def test_syn_tcp_client_instantiation(self):
        """Test sync tcp client."""
        client = ModbusTcpClient("127.0.0.1")
        self.assertNotEqual(client, None)

    @patch("pymodbus.client.tcp.select")
    def test_basic_syn_tcp_client(self, mock_select):
        """Test the basic methods for the tcp sync client"""
        # receive/send
        mock_select.select.return_value = [True]
        client = ModbusTcpClient("127.0.0.1")
        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(1, client.send(b"\x45"))
        self.assertEqual(b"\x45", client.recv(1))

        # connect/disconnect
        self.assertTrue(client.connect())
        client.close()

        # already closed socket
        client.socket = False
        client.close()

        self.assertEqual("ModbusTcpClient(127.0.0.1:502)", str(client))

    def test_tcp_client_is_socket_open(self):
        """Test the tcp client is_socket_open method"""
        client = ModbusTcpClient("127.0.0.1")
        self.assertFalse(client.is_socket_open())

    def test_tcp_client_send(self):
        """Test the tcp client send method"""
        client = ModbusTcpClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.send(None),
        )

        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(4, client.send("1234"))

    @patch("pymodbus.client.tcp.time")
    @patch("pymodbus.client.tcp.select")
    def test_tcp_client_recv(self, mock_select, mock_time):
        """Test the tcp client receive method"""
        mock_select.select.return_value = [True]
        mock_time.time.side_effect = count()
        client = ModbusTcpClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )
        client.socket = mockSocket()
        self.assertEqual(b"", client.recv(0))
        client.socket.mock_store(b"\x00" * 4)
        self.assertEqual(b"\x00" * 4, client.recv(4))

        mock_socket = MagicMock()
        mock_socket.recv.side_effect = iter([b"\x00", b"\x01", b"\x02"])
        client.socket = mock_socket
        client.params.timeout = 3
        self.assertEqual(b"\x00\x01\x02", client.recv(3))
        mock_socket.recv.side_effect = iter([b"\x00", b"\x01", b"\x02"])
        self.assertEqual(b"\x00\x01", client.recv(2))
        mock_select.select.return_value = [False]
        self.assertEqual(b"", client.recv(2))
        client.socket = mockSocket()
        client.socket.mock_store(b"\x00")
        mock_select.select.return_value = [True]
        self.assertIn(b"\x00", client.recv(None))

        mock_socket = MagicMock()
        mock_socket.recv.return_value = b""
        client.socket = mock_socket
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )

        mock_socket.recv.side_effect = iter([b"\x00", b"\x01", b"\x02", b""])
        client.socket = mock_socket
        self.assertEqual(b"\x00\x01\x02", client.recv(1024))

    def test_tcp_client_repr(self):
        """Test tcp client."""
        client = ModbusTcpClient("127.0.0.1")
        rep = (
            f"<{client.__class__.__name__} at {hex(id(client))} socket={client.socket}, "
            f"ipaddr={client.params.host}, port={client.params.port}, timeout={client.params.timeout}>"
        )
        self.assertEqual(repr(client), rep)

    def test_tcp_client_register(self):
        """Test tcp client."""

        class CustomRequest:  # pylint: disable=too-few-public-methods
            """Dummy custom request."""

            function_code = 79

        client = ModbusTcpClient("127.0.0.1")
        client.framer = Mock()
        client.register(CustomRequest)
        self.assertTrue(client.framer.decoder.register.called_once_with(CustomRequest))

    # -----------------------------------------------------------------------#
    # Test TLS Client
    # -----------------------------------------------------------------------#

    def test_tls_sslctx_provider(self):
        """Test that sslctx_provider() produce SSLContext correctly"""
        with patch.object(ssl.SSLContext, "load_cert_chain") as mock_method:
            sslctx1 = sslctx_provider(certfile="cert.pem")
            self.assertIsNotNone(sslctx1)
            self.assertEqual(type(sslctx1), ssl.SSLContext)
            self.assertEqual(mock_method.called, False)

            sslctx2 = sslctx_provider(keyfile="key.pem")
            self.assertIsNotNone(sslctx2)
            self.assertEqual(type(sslctx2), ssl.SSLContext)
            self.assertEqual(mock_method.called, False)

            sslctx3 = sslctx_provider(certfile="cert.pem", keyfile="key.pem")
            self.assertIsNotNone(sslctx3)
            self.assertEqual(type(sslctx3), ssl.SSLContext)
            self.assertEqual(mock_method.called, True)

            sslctx_old = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            sslctx_new = sslctx_provider(sslctx=sslctx_old)
            self.assertEqual(sslctx_new, sslctx_old)

    def test_syn_tls_client_instantiation(self):
        """Test sync tls client."""
        # default SSLContext
        client = ModbusTlsClient("127.0.0.1")
        self.assertNotEqual(client, None)
        self.assertIsInstance(client.framer, ModbusTlsFramer)
        self.assertTrue(client.sslctx)

    @patch("pymodbus.client.tcp.select")
    def test_basic_syn_tls_client(self, mock_select):
        """Test the basic methods for the tls sync client"""
        # receive/send
        mock_select.select.return_value = [True]
        client = ModbusTlsClient("localhost")
        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(1, client.send(b"\x45"))
        self.assertEqual(b"\x45", client.recv(1))

        # connect/disconnect
        self.assertTrue(client.connect())
        client.close()

        # already closed socket
        client.socket = False
        client.close()
        self.assertEqual("ModbusTlsClient(localhost:802)", str(client))

        client = ModbusTcpClient("127.0.0.1")
        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(1, client.send(b"\x45"))
        self.assertEqual(b"\x45", client.recv(1))

    def test_tls_client_send(self):
        """Test the tls client send method"""
        client = ModbusTlsClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.send(None),
        )

        client.socket = mockSocket()
        self.assertEqual(0, client.send(None))
        self.assertEqual(4, client.send("1234"))

    @patch("pymodbus.client.tcp.time")
    @patch("pymodbus.client.tcp.select")
    def test_tls_client_recv(self, mock_select, mock_time):
        """Test the tls client receive method"""
        mock_select.select.return_value = [True]
        client = ModbusTlsClient("127.0.0.1")
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )

        mock_time.time.side_effect = count()

        client.socket = mockSocket()
        client.socket.mock_store(b"\x00" * 4)
        self.assertEqual(b"", client.recv(0))
        self.assertEqual(b"\x00" * 4, client.recv(4))

        client.params.timeout = 2
        client.socket.mock_store(b"\x00")
        self.assertIn(b"\x00", client.recv(None))

        # client.socket = mockSocket()
        # client.socket.recv.side_effect = iter([b"\x00", b"\x01", b"\x02"])
        # client.params.timeout = 3
        # self.assertEqual(
        #     b"\x00\x01\x02", client.recv(3)
        # )
        # client.socket.recv.side_effect = iter([b"\x00", b"\x01", b"\x02"])
        # self.assertEqual(
        #     b"\x00\x01", client.recv(2)
        # )

    def test_tls_client_repr(self):
        """Test tls client."""
        client = ModbusTlsClient("127.0.0.1")
        rep = (
            f"<{client.__class__.__name__} at {hex(id(client))} socket={client.socket}, "
            f"ipaddr={client.params.host}, port={client.params.port}, sslctx={client.sslctx}, "
            f"timeout={client.params.timeout}>"
        )
        self.assertEqual(repr(client), rep)

    def test_tls_client_register(self):
        """Test tls client."""

        class CustomeRequest:  # pylint: disable=too-few-public-methods
            """Dummy custom request."""

            function_code = 79

        client = ModbusTlsClient("127.0.0.1")
        client.framer = Mock()
        client.register(CustomeRequest)
        self.assertTrue(client.framer.decoder.register.called_once_with(CustomeRequest))

    # -----------------------------------------------------------------------#
    # Test Serial Client
    # -----------------------------------------------------------------------#
    def test_sync_serial_client_instantiation(self):
        """Test sync serial client."""
        client = ModbusSerialClient("/dev/null")
        self.assertNotEqual(client, None)
        self.assertTrue(
            isinstance(
                ModbusSerialClient("/dev/null", framer=ModbusAsciiFramer).framer,
                ModbusAsciiFramer,
            )
        )
        self.assertTrue(
            isinstance(
                ModbusSerialClient("/dev/null", framer=ModbusRtuFramer).framer,
                ModbusRtuFramer,
            )
        )
        self.assertTrue(
            isinstance(
                ModbusSerialClient("/dev/null", framer=ModbusBinaryFramer).framer,
                ModbusBinaryFramer,
            )
        )
        self.assertTrue(
            isinstance(
                ModbusSerialClient("/dev/null", framer=ModbusSocketFramer).framer,
                ModbusSocketFramer,
            )
        )

    def test_sync_serial_rtu_client_timeouts(self):
        """Test sync serial rtu."""
        client = ModbusSerialClient("/dev/null", framer=ModbusRtuFramer, baudrate=9600)
        self.assertEqual(client.silent_interval, round((3.5 * 11 / 9600), 6))
        client = ModbusSerialClient("/dev/null", framer=ModbusRtuFramer, baudrate=38400)
        self.assertEqual(client.silent_interval, round((1.75 / 1000), 6))

    @patch("serial.Serial")
    def test_basic_sync_serial_client(self, mock_serial):
        """Test the basic methods for the serial sync client."""
        # receive/send
        mock_serial.in_waiting = 0
        mock_serial.write = lambda x: len(x)  # pylint: disable=unnecessary-lambda

        mock_serial.read = lambda size: b"\x00" * size
        client = ModbusSerialClient("/dev/null")
        client.socket = mock_serial
        client.state = 0
        self.assertEqual(0, client.send(None))
        client.state = 0
        self.assertEqual(1, client.send(b"\x00"))
        self.assertEqual(b"\x00", client.recv(1))

        # connect/disconnect
        self.assertTrue(client.connect())
        client.close()

        # rtu connect/disconnect
        rtu_client = ModbusSerialClient(
            "/dev/null", framer=ModbusRtuFramer, strict=True
        )
        self.assertTrue(rtu_client.connect())
        self.assertEqual(
            rtu_client.socket.interCharTimeout, rtu_client.inter_char_timeout
        )
        rtu_client.close()
        self.assertTrue("baud[19200])" in str(client))

        # already closed socket
        client.socket = False
        client.close()

    def test_serial_client_connect(self):
        """Test the serial client connection method"""
        with patch.object(serial, "Serial") as mock_method:
            mock_method.return_value = MagicMock()
            client = ModbusSerialClient("/dev/null")
            self.assertTrue(client.connect())

        with patch.object(serial, "Serial") as mock_method:
            mock_method.side_effect = serial.SerialException()
            client = ModbusSerialClient("/dev/null")
            self.assertFalse(client.connect())

    @patch("serial.Serial")
    def test_serial_client_is_socket_open(self, mock_serial):
        """Test the serial client is_socket_open method"""
        client = ModbusSerialClient("/dev/null")
        self.assertFalse(client.is_socket_open())
        client.socket = mock_serial
        self.assertTrue(client.is_socket_open())

    @patch("serial.Serial")
    def test_serial_client_send(self, mock_serial):
        """Test the serial client send method"""
        mock_serial.in_waiting = None
        mock_serial.write = lambda x: len(x)  # pylint: disable=unnecessary-lambda
        client = ModbusSerialClient("/dev/null")
        self.assertRaises(
            ConnectionException,
            lambda: client.send(None),
        )
        # client.connect()
        client.socket = mock_serial
        client.state = 0
        self.assertEqual(0, client.send(None))
        client.state = 0
        self.assertEqual(4, client.send("1234"))

    @patch("serial.Serial")
    def test_serial_client_cleanup_buffer_before_send(self, mock_serial):
        """Test the serial client send method"""
        mock_serial.in_waiting = 4
        mock_serial.read = lambda x: b"1" * x
        mock_serial.write = lambda x: len(x)  # pylint: disable=unnecessary-lambda
        client = ModbusSerialClient("/dev/null")
        self.assertRaises(
            ConnectionException,
            lambda: client.send(None),
        )
        # client.connect()
        client.socket = mock_serial
        client.state = 0
        self.assertEqual(0, client.send(None))
        client.state = 0
        self.assertEqual(4, client.send("1234"))

    def test_serial_client_recv(self):
        """Test the serial client receive method"""
        client = ModbusSerialClient("/dev/null")
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )

        client.socket = mockSocket()
        self.assertEqual(b"", client.recv(0))
        client.socket.mock_store(b"\x00" * 4)
        self.assertEqual(b"\x00" * 4, client.recv(4))
        client.socket = MagicMock()
        client.socket.read.return_value = b""
        self.assertEqual(b"", client.recv(None))
        client.socket.timeout = 0
        self.assertEqual(b"", client.recv(0))
        client.params.timeout = None
        self.assertEqual(b"", client.recv(None))

    def test_serial_client_repr(self):
        """Test serial client."""
        client = ModbusSerialClient("/dev/null")
        rep = (
            f"<{client.__class__.__name__} at {hex(id(client))} socket={client.socket}, "
            f"framer={client.framer}, timeout={client.params.timeout}>"
        )
        self.assertEqual(repr(client), rep)



"""RTU framer."""
# pylint: disable=missing-type-doc
import logging
import struct
import time

from pymodbus.exceptions import (
    InvalidMessageReceivedException,
    ModbusIOException,
)
from pymodbus.framer import BYTE_ORDER, FRAME_HEADER, ModbusFramer
from pymodbus.utilities import (
    ModbusTransactionState,
    checkCRC,
    computeCRC,
    hexlify_packets,
)


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)

RTU_FRAME_HEADER = BYTE_ORDER + FRAME_HEADER


# --------------------------------------------------------------------------- #
# Modbus RTU Message
# --------------------------------------------------------------------------- #
class ModbusRtuFramer(ModbusFramer):
    """Modbus RTU Frame controller.

        [ Start Wait ] [Address ][ Function Code] [ Data ][ CRC ][  End Wait  ]
          3.5 chars     1b         1b               Nb      2b      3.5 chars

    Wait refers to the amount of time required to transmit at least x many
    characters.  In this case it is 3.5 characters.  Also, if we receive a
    wait of 1.5 characters at any point, we must trigger an error message.
    Also, it appears as though this message is little endian. The logic is
    simplified as the following::

        block-on-read:
            read until 3.5 delay
            check for errors
            decode

    The following table is a listing of the baud wait times for the specified
    baud rates::

        ------------------------------------------------------------------
         Baud  1.5c (18 bits)   3.5c (38 bits)
        ------------------------------------------------------------------
         1200   13333.3 us       31666.7 us
         4800    3333.3 us        7916.7 us
         9600    1666.7 us        3958.3 us
        19200     833.3 us        1979.2 us
        38400     416.7 us         989.6 us
        ------------------------------------------------------------------
        1 Byte = start + 8 bits + parity + stop = 11 bits
        (1/Baud)(bits) = delay seconds
    """

    method = "rtu"

    def __init__(self, decoder, client=None):
        """Initialize a new instance of the framer.

        :param decoder: The decoder factory implementation to use
        """
        self._buffer = b""
        self._header = {"uid": 0x00, "len": 0, "crc": b"\x00\x00"}
        self._hsize = 0x01
        self._end = b"\x0d\x0a"
        self._min_frame_size = 4
        self.decoder = decoder
        self.client = client

    # ----------------------------------------------------------------------- #
    # Private Helper Functions
    # ----------------------------------------------------------------------- #
    def decode_data(self, data):
        """Decode data."""
        if len(data) > self._hsize:
            uid = int(data[0])
            fcode = int(data[1])
            return {"unit": uid, "fcode": fcode}
        return {}

    def checkFrame(self):
        """Check if the next frame is available.

        Return True if we were successful.

        1. Populate header
        2. Discard frame if UID does not match
        """
        try:
            self.populateHeader()
            frame_size = self._header["len"]
            data = self._buffer[: frame_size - 2]
            crc = self._header["crc"]
            crc_val = (int(crc[0]) << 8) + int(crc[1])
            return checkCRC(data, crc_val)
        except (IndexError, KeyError, struct.error):
            return False

    def advanceFrame(self):
        """Skip over the current framed message.

        This allows us to skip over the current message after we have processed
        it or determined that it contains an error. It also has to reset the
        current frame header handle
        """
        self._buffer = self._buffer[self._header["len"] :]
        _logger.debug("Frame advanced, resetting header!!")
        self._header = {"uid": 0x00, "len": 0, "crc": b"\x00\x00"}

    def resetFrame(self):  # pylint: disable=invalid-name
        """Reset the entire message frame.

        This allows us to skip over errors that may be in the stream.
        It is hard to know if we are simply out of sync or if there is
        an error in the stream as we have no way to check the start or
        end of the message (python just doesn"t have the resolution to
        check for millisecond delays).
        """
        txt = (
            f"Resetting frame - Current Frame in "
            f"buffer - {hexlify_packets(self._buffer)}"
        )
        _logger.debug(txt)
        self._buffer = b""
        self._header = {"uid": 0x00, "len": 0, "crc": b"\x00\x00"}

    def isFrameReady(self):
        """Check if we should continue decode logic.

        This is meant to be used in a while loop in the decoding phase to let
        the decoder know that there is still data in the buffer.

        :returns: True if ready, False otherwise
        """
        size = self._header.get("len", 0)
        if not size and len(self._buffer) > self._hsize:
            try:
                # Frame is ready only if populateHeader() successfully
                # populates crc field which finishes RTU frame otherwise,
                # if buffer is not yet long enough, populateHeader() raises IndexError
                size = self.populateHeader()
            except IndexError:
                return False

        return len(self._buffer) >= size if size > 0 else False

    def populateHeader(self, data=None):  # pylint: disable=invalid-name
        """Try to set the headers `uid`, `len` and `crc`.

        This method examines `self._buffer` and writes meta
        information into `self._header`.

        Beware that this method will raise an IndexError if
        `self._buffer` is not yet long enough.
        """
        data = data if data is not None else self._buffer
        self._header["uid"] = int(data[0])
        size = self.get_expected_response_length(data)
        self._header["len"] = size

        if len(data) < size:
            # crc yet not available
            raise IndexError
        self._header["crc"] = data[size - 2 : size]
        return size

    def addToFrame(self, message):
        """Add the received data to the buffer handle.

        :param message: The most recent packet
        """
        self._buffer += message

    def getFrame(self):
        """Get the next frame from the buffer.

        :returns: The frame data or ""
        """
        start = self._hsize
        end = self._header["len"] - 2
        buffer = self._buffer[start:end]
        if end > 0:
            txt = f"Getting Frame - {hexlify_packets(buffer)}"
            _logger.debug(txt)
            return buffer
        return b""

    def populateResult(self, result):
        """Populate the modbus result header.

        The serial packets do not have any header information
        that is copied.

        :param result: The response packet
        """
        result.unit_id = self._header["uid"]
        result.transaction_id = self._header["uid"]

    # ----------------------------------------------------------------------- #
    # Public Member Functions
    # ----------------------------------------------------------------------- #
    def processIncomingPacket(
        self, data, callback, unit, **kwargs
    ):  # pylint: disable=arguments-differ
        """Process new packet pattern.

        This takes in a new request packet, adds it to the current
        packet stream, and performs framing on it. That is, checks
        for complete messages, and once found, will process all that
        exist.  This handles the case when we read N + 1 or 1 // N
        messages at a time instead of 1.

        The processed and decoded messages are pushed to the callback
        function to process and send.

        :param data: The new packet data
        :param callback: The function to send results to
        :param unit: Process if unit id matches, ignore otherwise (could be a
               list of unit ids (server) or single unit id(client/server)
        :param kwargs:
        """
        if not isinstance(unit, (list, tuple)):
            unit = [unit]
        self.addToFrame(data)
        single = kwargs.get("single", False)
        while True:
            if self.isFrameReady():
                if self.checkFrame():
                    if self._validate_unit_id(unit, single):
                        self._process(callback)
                    else:
                        header_txt = self._header["uid"]
                        txt = f"Not a valid unit id - {header_txt}, ignoring!!"
                        _logger.debug(txt)
                        self.resetFrame()
                        break
                else:
                    _logger.debug("Frame check failed, ignoring!!")
                    self.resetFrame()
                    break
            else:
                txt = f"Frame - [{data}] not ready"
                _logger.debug(txt)
                break

    def buildPacket(self, message):
        """Create a ready to send modbus packet.

        :param message: The populated request/response to send
        """
        data = message.encode()
        packet = (
            struct.pack(RTU_FRAME_HEADER, message.unit_id, message.function_code) + data
        )
        packet += struct.pack(">H", computeCRC(packet))
        # Ensure that transaction is actually the unit id for serial comms
        message.transaction_id = message.unit_id
        return packet

    def sendPacket(self, message):
        """Send packets on the bus with 3.5char delay between frames.

        :param message: Message to be sent over the bus
        :return:
        """
        start = time.time()
        timeout = start + self.client.params.timeout
        while self.client.state != ModbusTransactionState.IDLE:
            if self.client.state == ModbusTransactionState.TRANSACTION_COMPLETE:
                timestamp = round(time.time(), 6)
                txt = (
                    f"Changing state to IDLE - Last Frame End - {self.client.last_frame_end}, "
                    f"Current Time stamp - {timestamp}"
                )
                _logger.debug(txt)

                if self.client.last_frame_end:
                    idle_time = self.client.idle_time()
                    if round(timestamp - idle_time, 6) <= self.client.silent_interval:
                        txt = (
                            f"Waiting for 3.5 char before next "
                            f"send - {self.client.silent_interval * 1000} ms"
                        )
                        _logger.debug(txt)
                        time.sleep(self.client.silent_interval)
                else:
                    # Recovering from last error ??
                    time.sleep(self.client.silent_interval)
                self.client.state = ModbusTransactionState.IDLE
            elif self.client.state == ModbusTransactionState.RETRYING:
                # Simple lets settle down!!!
                # To check for higher baudrates
                time.sleep(self.client.params.timeout)
                break
            elif time.time() > timeout:
                _logger.debug(
                    "Spent more time than the read time out, "
                    "resetting the transaction to IDLE"
                )
                self.client.state = ModbusTransactionState.IDLE
            else:
                _logger.debug("Sleeping")
                time.sleep(self.client.silent_interval)
        size = self.client.send(message)
        self.client.last_frame_end = round(time.time(), 6)
        return size

    def recvPacket(self, size):
        """Receive packet from the bus with specified len.

        :param size: Number of bytes to read
        :return:
        """
        result = self.client.recv(size)
        self.client.last_frame_end = round(time.time(), 6)
        return result

    def _process(self, callback, error=False):
        """Process incoming packets irrespective error condition."""
        data = self.getRawFrame() if error else self.getFrame()
        if (result := self.decoder.decode(data)) is None:
            raise ModbusIOException("Unable to decode request")
        if error and result.function_code < 0x80:
            raise InvalidMessageReceivedException(result)
        self.populateResult(result)
        self.advanceFrame()
        callback(result)  # defer or push to a thread?

    def getRawFrame(self):  # pylint: disable=invalid-name
        """Return the complete buffer."""
        txt = f"Getting Raw Frame - {hexlify_packets(self._buffer)}"
        _logger.debug(txt)
        return self._buffer

    def get_expected_response_length(self, data):
        """Get the expected response length.

        :param data: Message data read so far
        :raises IndexError: If not enough data to read byte count
        :return: Total frame size
        """
        func_code = int(data[1])
        pdu_class = self.decoder.lookupPduClass(func_code)
        return pdu_class.calculateRtuFrameSize(data)


# __END__



"""Modbus client async UDP communication."""
import asyncio
import functools
import logging
import socket
import typing

from pymodbus.client.base import ModbusBaseClient, ModbusClientProtocol
from pymodbus.constants import Defaults
from pymodbus.framer import ModbusFramer
from pymodbus.framer.socket_framer import ModbusSocketFramer
from pymodbus.exceptions import ConnectionException

_logger = logging.getLogger(__name__)

DGRAM_TYPE = socket.SOCK_DGRAM


class AsyncModbusUdpClient(ModbusBaseClient):
    """**AsyncModbusUdpClient**.

    :param host: Host IP address or host name
    :param port: (optional) Port used for communication.
    :param framer: (optional) Framer class.
    :param source_address: (optional) source address of client,
    :param kwargs: (optional) Experimental parameters

    Example::

        from pymodbus.client import AsyncModbusUdpClient

        async def run():
            client = AsyncModbusUdpClient("localhost")

            await client.connect()
            ...
            await client.close()
    """

    def __init__(
        self,
        host: str,
        port: int = Defaults.UdpPort,
        framer: ModbusFramer = ModbusSocketFramer,
        source_address: typing.Tuple[str, int] = None,
        **kwargs: any,
    ) -> None:
        """Initialize Asyncio Modbus UDP Client."""
        self.protocol = None
        super().__init__(framer=framer, **kwargs)
        self.params.host = host
        self.params.port = port
        self.params.source_address = source_address

        self.loop = asyncio.get_event_loop()
        self.connected = False
        self.delay_ms = self.params.reconnect_delay
        self.reset_delay()

    async def connect(self):  # pylint: disable=invalid-overridden-method
        """Start reconnecting asynchronous udp client.

        :meta private:
        """
        # force reconnect if required:
        host = self.params.host
        await self.close()
        self.params.host = host

        # get current loop, if there are no loop a RuntimeError will be raised
        self.loop = asyncio.get_running_loop()

        txt = f"Connecting to {self.params.host}:{self.params.port}."
        _logger.debug(txt)

        # getaddrinfo returns a list of tuples
        # - [(family, type, proto, canonname, sockaddr),]
        # We want sockaddr which is a (ip, port) tuple
        # udp needs ip addresses, not hostnames
        # TBD: addrinfo = await self.loop.getaddrinfo(self.params.host, self.params.port, type=DGRAM_TYPE)
        # TBD: self.params.host, self.params.port = addrinfo[-1][-1]
        return await self._connect()

    async def close(self):  # pylint: disable=invalid-overridden-method
        """Stop connection and prevents reconnect.

        :meta private:
        """
        # prevent reconnect:
        self.params.host = None

        if self.connected and self.protocol and self.protocol.transport:
            self.protocol.transport.close()

    def _create_protocol(self, host=None, port=0):
        """Create initialized protocol instance with factory function."""
        protocol = ModbusClientProtocol(
            use_udp=True, framer=self.params.framer, **self.params.kwargs
        )
        protocol.params.host = host
        protocol.params.port = port
        protocol.factory = self
        return protocol

    async def _connect(self):
        """Connect."""
        _logger.debug("Connecting.")
        try:
            endpoint = await self.loop.create_datagram_endpoint(
                functools.partial(
                    self._create_protocol, host=self.params.host, port=self.params.port
                ),
                remote_addr=(self.params.host, self.params.port),
            )
            txt = f"Connected to {self.params.host}:{self.params.port}."
            _logger.info(txt)
            return endpoint
        except Exception as exc:  # pylint: disable=broad-except
            txt = f"Failed to connect: {exc}"
            _logger.warning(txt)
            asyncio.ensure_future(self._reconnect())

    def protocol_made_connection(self, protocol):
        """Notify successful connection.

        :meta private:
        """
        _logger.info("Protocol made connection.")
        if not self.connected:
            self.connected = True
            self.protocol = protocol
        else:
            _logger.error("Factory protocol connect callback called while connected.")

    def protocol_lost_connection(self, protocol):
        """Notify lost connection.

        :meta private:
        """
        if self.connected:
            _logger.info("Protocol lost connection.")
            if protocol is not self.protocol:
                _logger.error(
                    "Factory protocol callback called "
                    "from unexpected protocol instance."
                )

            self.connected = False
            self.protocol = None
            if self.params.host:
                asyncio.create_task(self._reconnect())
        else:
            _logger.error("Factory protocol connect callback called while connected.")

    async def _reconnect(self):
        """Reconnect."""
        txt = f"Waiting {self.delay_ms} ms before next connection attempt."
        _logger.debug(txt)
        await asyncio.sleep(self.delay_ms / 1000)
        self.delay_ms = 2 * self.delay_ms
        return await self._connect()


class ModbusUdpClient(ModbusBaseClient):
    """**ModbusUdpClient**.

    :param host: Host IP address or host name
    :param port: (optional) Port used for communication.
    :param framer: (optional) Framer class.
    :param source_address: (optional) source address of client,
    :param kwargs: (optional) Experimental parameters

    Example::

        from pymodbus.client import ModbusUdpClient

        async def run():
            client = ModbusUdpClient("localhost")

            client.connect()
            ...
            client.close()
    """

    def __init__(
        self,
        host: str,
        port: int = Defaults.UdpPort,
        framer: ModbusFramer = ModbusSocketFramer,
        source_address: typing.Tuple[str, int] = None,
        **kwargs: any,
    ) -> None:
        """Initialize Asyncio Modbus UDP Client."""
        super().__init__(framer=framer, **kwargs)
        self.params.host = host
        self.params.port = port
        self.params.source_address = source_address

        self.socket = None

    @property
    def connected(self):
        """Connect internal.

        :meta private:
        """
        return self.connect()

    def connect(self):
        """Connect to the modbus tcp server.

        :meta private:
        """
        if self.socket:
            return True
        try:
            family = ModbusUdpClient._get_address_family(self.params.host)
            self.socket = socket.socket(family, socket.SOCK_DGRAM)
            self.socket.settimeout(self.params.timeout)
        except socket.error as exc:
            txt = f"Unable to create udp socket {exc}"
            _logger.error(txt)
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection.

        :meta private:
        """
        self.socket = None

    def send(self, request):
        """Send data on the underlying socket.

        :meta private:
        """
        super().send(request)
        if not self.socket:
            raise ConnectionException(str(self))
        if request:
            return self.socket.sendto(request, (self.params.host, self.params.port))
        return 0

    def recv(self, size):
        """Read data from the underlying descriptor.

        :meta private:
        """
        super().recv(size)
        if not self.socket:
            raise ConnectionException(str(self))
        return self.socket.recvfrom(size)[0]

    def is_socket_open(self):
        """Check if socket is open.

        :meta private:
        """
        if self.socket:
            return True
        return self.connect()

    def __str__(self):
        """Build a string representation of the connection."""
        return f"ModbusUdpClient({self.params.host}:{self.params.port})"

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"ipaddr={self.params.host}, port={self.params.port}, timeout={self.params.timeout}>"
        )



#!/usr/bin/env python3
# pylint: disable=missing-raises-doc
"""An example of creating a fully implemented modbus server.

with read/write data as well as user configurable base data
"""
import logging
from optparse import OptionParser  # pylint: disable=deprecated-module
import pickle  # nosec

from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext
from pymodbus.server import StartTcpServer


# -------------------------------------------------------------------------- #
# Logging
# -------------------------------------------------------------------------- #
server_log = logging.getLogger("pymodbus.server")
protocol_log = logging.getLogger("pymodbus.protocol")
_logger = logging.getLogger(__name__)

# -------------------------------------------------------------------------- #
# Extra Global Functions
# -------------------------------------------------------------------------- #
# These are extra helper functions that don"t belong in a class
# -------------------------------------------------------------------------- #
# import getpass


def root_test():
    """Check to see if we are running as root"""
    return True  # removed for the time being as it isn"t portable
    # return getpass.getuser() == "root"


# -------------------------------------------------------------------------- #
# Helper Classes
# -------------------------------------------------------------------------- #


class ConfigurationException(Exception):
    """Exception for configuration error"""

    def __init__(self, string):
        """Initialize the ConfigurationException instance

        :param string: The message to append to the exception
        """
        Exception.__init__(self, string)
        self.string = string

    def __str__(self):
        """Build a representation of the object

        :returns: A string representation of the object
        """
        return f"Configuration Error: {self.string}"


class Configuration:  # pylint: disable=too-few-public-methods
    """Class used to parse configuration file and create and modbus datastore.

    The format of the configuration file is actually just a
    python pickle, which is a compressed memory dump from
    the scraper.
    """

    def __init__(self, config):
        """Try to load a configuration file.

        lets the file not found exception fall through

        :param config: The pickled datastore
        """
        try:
            self.file = open(config, "rb")  # pylint: disable=consider-using-with
        except Exception as exc:
            _logger.critical(str(exc))
            raise ConfigurationException(  # pylint: disable=raise-missing-from
                f"File not found {config}"
            )

    def parse(self):
        """Parse the config file and creates a server context"""
        handle = pickle.load(self.file)  # nosec
        try:  # test for existence, or bomb
            dsd = handle["di"]
            csd = handle["ci"]
            hsd = handle["hr"]
            isd = handle["ir"]
        except Exception:
            raise ConfigurationException(  # pylint: disable=raise-missing-from
                "Invalid Configuration"
            )
        slave = ModbusSlaveContext(d=dsd, c=csd, h=hsd, i=isd)
        return ModbusServerContext(slaves=slave)


# -------------------------------------------------------------------------- #
# Main start point
# -------------------------------------------------------------------------- #


def main():
    """Server launcher"""
    parser = OptionParser()
    parser.add_option(
        "-c", "--conf", help="The configuration file to load", dest="file"
    )
    parser.add_option(
        "-D",
        "--debug",
        help="Turn on to enable tracing",
        action="store_true",
        dest="debug",
        default=False,
    )
    (opt, _) = parser.parse_args()

    # enable debugging information
    if opt.debug:
        try:
            server_log.setLevel(logging.DEBUG)
            protocol_log.setLevel(logging.DEBUG)
        except Exception:  # pylint: disable=broad-except
            print("Logging is not supported on this system")

    # parse configuration file and run
    try:
        conf = Configuration(opt.file)
        StartTcpServer(context=conf.parse())
    except ConfigurationException as err:
        print(err)
        parser.print_help()


# -------------------------------------------------------------------------- #
# Main jumper
# -------------------------------------------------------------------------- #


if __name__ == "__main__":
    if root_test():
        main()
    else:
        print("This script must be run as root!")



#!/usr/bin/env python3
"""Pymodbus Synchronous Client extended calls rxample.

This example uses client_sync.py to handle connection, and have the same options.

The example shows how to use the synchronous modbus client
implementation from pymodbus to perform the extended portions of the
modbus protocol.

If you are performing a request that is not available in the client
mixin, you have to perform the request like this instead:

from pymodbus.diag_message import ClearCountersRequest
from pymodbus.diag_message import ClearCountersResponse

request  = ClearCountersRequest()
response = client.execute(request)
if isinstance(response, ClearCountersResponse):
    ... do something with the response


The corresponding server must be started before e.g. as:
    python3 server_sync.py
"""
import logging

from examples.client_sync import run_sync_client, setup_sync_client

from pymodbus.diag_message import (
    ChangeAsciiInputDelimiterRequest,
    ClearCountersRequest,
    ClearOverrunCountRequest,
    ForceListenOnlyModeRequest,
    GetClearModbusPlusRequest,
    RestartCommunicationsOptionRequest,
    ReturnBusCommunicationErrorCountRequest,
    ReturnBusExceptionErrorCountRequest,
    ReturnDiagnosticRegisterRequest,
    ReturnIopOverrunCountRequest,
    ReturnQueryDataRequest,
    ReturnSlaveBusCharacterOverrunCountRequest,
    ReturnSlaveBusyCountRequest,
    ReturnSlaveMessageCountRequest,
    ReturnSlaveNAKCountRequest,
    ReturnSlaveNoResponseCountRequest,
)
from pymodbus.mei_message import ReadDeviceInformationRequest
from pymodbus.other_message import (
    GetCommEventCounterRequest,
    GetCommEventLogRequest,
    ReadExceptionStatusRequest,
    ReportSlaveIdRequest,
)


UNIT = 0x01


def _execute_information_requests(client):
    """Execute extended information requests."""
    _logger.info("### Running ReadDeviceInformationRequest")
    rr = client.execute(ReadDeviceInformationRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReportSlaveIdRequest")
    rr = client.execute(ReportSlaveIdRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert rr.status  # test that the status is ok

    _logger.info("Running ReadExceptionStatusRequest")
    rr = client.execute(ReadExceptionStatusRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert not rr.status  # test the status code

    _logger.info("Running GetCommEventCounterRequest")
    rr = client.execute(GetCommEventCounterRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert rr.status  # test the status code
    assert not rr.count  # test the count returned

    _logger.info("Running GetCommEventLogRequest")
    rr = client.execute(GetCommEventLogRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert rr.status  # test the status code
    assert not rr.event_count  # test the number of events
    assert not rr.message_count  # test the number of messages
    assert not rr.events  # test the number of events


def _execute_diagnostic_requests(client):
    """Execute extended diagnostic requests."""
    _logger.info("### Running ReturnQueryDataRequest")
    rr = client.execute(ReturnQueryDataRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert not rr.message[0]  # test the resulting message

    _logger.info("Running RestartCommunicationsOptionRequest")
    rr = client.execute(RestartCommunicationsOptionRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK
    assert not rr.message[0]  # test the resulting message

    _logger.info("Running ReturnDiagnosticRegisterRequest")
    rr = client.execute(ReturnDiagnosticRegisterRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ChangeAsciiInputDelimiterRequest")
    rr = client.execute(ChangeAsciiInputDelimiterRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ForceListenOnlyModeRequest")
    _logger.info("NOT WORKING")
    _logger.info(str(ForceListenOnlyModeRequest))
    # rr = client.execute(
    #    ForceListenOnlyModeRequest(unit=UNIT)
    # )  # does not send a response
    # assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ClearCountersRequest")
    rr = client.execute(ClearCountersRequest())
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnBusCommunicationErrorCountRequest")
    rr = client.execute(ReturnBusCommunicationErrorCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnBusExceptionErrorCountRequest")
    rr = client.execute(ReturnBusExceptionErrorCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnSlaveMessageCountRequest")
    rr = client.execute(ReturnSlaveMessageCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnSlaveNoResponseCountRequest")
    rr = client.execute(ReturnSlaveNoResponseCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnSlaveNAKCountRequest")
    rr = client.execute(ReturnSlaveNAKCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnSlaveBusyCountRequest")
    rr = client.execute(ReturnSlaveBusyCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnSlaveBusCharacterOverrunCountRequest")
    rr = client.execute(ReturnSlaveBusCharacterOverrunCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ReturnIopOverrunCountRequest")
    rr = client.execute(ReturnIopOverrunCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running ClearOverrunCountRequest")
    rr = client.execute(ClearOverrunCountRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK

    _logger.info("Running GetClearModbusPlusRequest")
    rr = client.execute(GetClearModbusPlusRequest(unit=UNIT))
    assert rr and not rr.isError()  # test that calls was OK


def run_sync_ext_calls(client):
    """Demonstrate basic read/write calls."""
    _execute_information_requests(client)
    _execute_diagnostic_requests(client)


# --------------------------------------------------------------------------- #
# Extra code, to allow commandline parameters instead of changing the code
# --------------------------------------------------------------------------- #
FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
logging.basicConfig(format=FORMAT)
_logger = logging.getLogger()


if __name__ == "__main__":
    testclient = setup_sync_client()
    run_sync_client(testclient, modbus_calls=run_sync_ext_calls)



"""Repl server main."""
from __future__ import annotations

import asyncio
from enum import Enum
import json
import logging
from pathlib import Path
from typing import List

import typer

from pymodbus.framer.socket_framer import ModbusSocketFramer
from pymodbus.repl.server.cli import run_repl
from pymodbus.server.reactive.default_config import DEFAULT_CONFIG
from pymodbus.server.reactive.main import (
    DEFAULT_FRAMER,
    DEFUALT_HANDLERS,
    ReactiveServer,
)


CANCELLED_ERROR = asyncio.exceptions.CancelledError

_logger = logging.getLogger(__name__)

CONTEXT_SETTING = {"allow_extra_args": True, "ignore_unknown_options": True}

# TBD class ModbusServerConfig:


class ModbusServerTypes(str, Enum):
    """Server types."""

    # ["tcp", "serial", "tls", "udp"]
    tcp = "tcp"  # pylint: disable=invalid-name
    serial = "serial"  # pylint: disable=invalid-name
    tls = "tls"  # pylint: disable=invalid-name
    udp = "udp"  # pylint: disable=invalid-name


class ModbusFramerTypes(str, Enum):
    """Framer types."""

    # ["socket", "rtu", "tls", "ascii", "binary"]
    socket = "socket"  # pylint: disable=invalid-name
    rtu = "rtu"  # pylint: disable=invalid-name
    tls = "tls"  # pylint: disable=invalid-name
    ascii = "ascii"  # pylint: disable=invalid-name
    binary = "binary"  # pylint: disable=invalid-name


def _completer(incomplete: str, valid_values: List[str]) -> List[str]:
    """Complete value."""
    completion = []
    for name in valid_values:
        if name.startswith(incomplete):
            completion.append(name)
    return completion


def framers(incomplete: str) -> List[str]:
    """Return an autocompleted list of supported clouds."""
    _framers = ["socket", "rtu", "tls", "ascii", "binary"]
    return _completer(incomplete, _framers)


def servers(incomplete: str) -> List[str]:
    """Return an autocompleted list of supported clouds."""
    _servers = ["tcp", "serial", "tls", "udp"]
    return _completer(incomplete, _servers)


app = typer.Typer(
    no_args_is_help=True,
    context_settings=CONTEXT_SETTING,
    help="Reactive modebus server",
)


@app.callback()
def server(
    ctx: typer.Context,
    host: str = typer.Option("localhost", "--host", help="Host address"),
    web_port: int = typer.Option(8080, "--web-port", help="Web app port"),
    broadcast_support: bool = typer.Option(
        False, "-b", help="Support broadcast messages"
    ),
    repl: bool = typer.Option(True, help="Enable/Disable repl for server"),
    verbose: bool = typer.Option(
        False, help="Run with debug logs enabled for pymodbus"
    ),
):
    """Run server code."""
    FORMAT = (  # pylint: disable=invalid-name
        "%(asctime)-15s %(threadName)-15s"
        " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
    )
    pymodbus_logger = logging.getLogger("pymodbus")
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger(__name__)
    if verbose:
        pymodbus_logger.setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
    else:
        pymodbus_logger.setLevel(logging.ERROR)
        logger.setLevel(logging.ERROR)

    ctx.obj = {
        "repl": repl,
        "host": host,
        "web_port": web_port,
        "broadcast": broadcast_support,
    }


@app.command("run", context_settings=CONTEXT_SETTING)
def run(
    ctx: typer.Context,
    modbus_server: str = typer.Option(
        ModbusServerTypes.tcp,
        "--modbus-server",
        "-s",
        case_sensitive=False,
        autocompletion=servers,
        help="Modbus Server",
    ),
    modbus_framer: str = typer.Option(
        ModbusFramerTypes.socket,
        "--framer",
        "-f",
        case_sensitive=False,
        autocompletion=framers,
        help="Modbus framer to use",
    ),
    modbus_port: str = typer.Option("5020", "--modbus-port", "-p", help="Modbus port"),
    modbus_unit_id: List[int] = typer.Option(
        None, "--unit-id", "-u", help="Supported Modbus unit id's"
    ),
    modbus_config: Path = typer.Option(
        None, help="Path to additional modbus server config"
    ),
    randomize: int = typer.Option(
        0,
        "--random",
        "-r",
        help="Randomize every `r` reads. 0=never, 1=always,2=every-second-read"
        ", and so on. Applicable IR and DI.",
    ),
):
    """Run Reactive Modbus server.

    Exposing REST endpoint for response manipulation.
    """
    repl = ctx.obj.pop("repl")
    # TBD extra_args = ctx.args
    web_app_config = ctx.obj
    loop = asyncio.get_event_loop()
    framer = DEFAULT_FRAMER.get(modbus_framer, ModbusSocketFramer)
    if modbus_config:
        with open(modbus_config) as my_file:  # pylint: disable=unspecified-encoding
            modbus_config = json.load(my_file)
    else:
        modbus_config = DEFAULT_CONFIG

    modbus_config = modbus_config.get(modbus_server, {})

    if modbus_server != "serial":
        modbus_port = int(modbus_port)
        handler = modbus_config.pop("handler", "ModbusConnectedRequestHandler")
    else:
        handler = modbus_config.pop("handler", "ModbusSingleRequestHandler")
    handler = DEFUALT_HANDLERS.get(handler.strip())

    modbus_config["handler"] = handler
    modbus_config["randomize"] = randomize
    app = ReactiveServer.factory(
        modbus_server,
        framer,
        modbus_port=modbus_port,
        unit=modbus_unit_id,
        loop=loop,
        single=False,
        **web_app_config,
        **modbus_config,
    )
    try:
        if repl:
            loop.run_until_complete(app.run_async())

            loop.run_until_complete(run_repl(app))
            loop.run_forever()
        else:
            app.run()

    except CANCELLED_ERROR:
        print("Done!!!!!")


if __name__ == "__main__":
    app()



"""Modbus Client Common."""
import logging
from typing import List, Union

import pymodbus.bit_read_message as pdu_bit_read
import pymodbus.bit_write_message as pdu_bit_write
from pymodbus.constants import Defaults
import pymodbus.diag_message as pdu_diag
import pymodbus.other_message as pdu_other_msg
from pymodbus.pdu import ModbusRequest, ModbusResponse
import pymodbus.register_read_message as pdu_reg_read
import pymodbus.register_write_message as pdu_req_write
from pymodbus.utilities import ModbusTransactionState


_logger = logging.getLogger(__name__)


class ModbusClientMixin:  # pylint: disable=too-many-public-methods
    """**ModbusClientMixin**.

    Simple modbus message call::

        response = client.read_coils(1, 10)
        # or
        response = await client.read_coils(1, 10)

    Advanced modbus message call::

        request = ReadCoilsRequest(1,10)
        response = client.execute(request)
        # or
        request = ReadCoilsRequest(1,10)
        response = await client.execute(request)

    .. tip::
        All methods can be used directly (synchronous) or with await <method>
        depending on the instantiated client.
    """

    state = ModbusTransactionState.IDLE
    last_frame_end = 0
    silent_interval = 0

    def execute(self, request: ModbusRequest) -> ModbusResponse:
        """Execute request.

        :param request: Request to send
        :raises ModbusException:
        """
        return request

    def read_coils(
        self,
        address: int,
        count: int = Defaults.Count,
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_bit_read.ReadCoilsResponse:
        """Read coils (function code 0x01).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_bit_read.ReadCoilsRequest(address, count, slave, **kwargs)
        return self.execute(request)

    def read_discrete_inputs(
        self,
        address: int,
        count: int = Defaults.Count,
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_bit_read.ReadDiscreteInputsResponse:
        """Read discrete inputs (function code 0x02).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_bit_read.ReadDiscreteInputsRequest(
            address, count, slave, **kwargs
        )
        return self.execute(request)

    def read_holding_registers(
        self,
        address: int,
        count: int = Defaults.Count,
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_reg_read.ReadHoldingRegistersResponse:
        """Read holding registers (function code 0x03).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_reg_read.ReadHoldingRegistersRequest(
            address, count, slave, **kwargs
        )
        return self.execute(request)

    def read_input_registers(
        self,
        address: int,
        count: int = Defaults.Count,
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_reg_read.ReadInputRegistersResponse:
        """Read input registers (function code 0x04).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_reg_read.ReadInputRegistersRequest(
            address, count, slave, **kwargs
        )
        return self.execute(request)

    def write_coil(
        self, address: int, value: bool, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_bit_write.WriteSingleCoilResponse:
        """Write single coil (function code 0x05).

        :param address: Start address to read from
        :param value: Boolean to write
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_bit_write.WriteSingleCoilRequest(address, value, slave, **kwargs)
        return self.execute(request)

    def write_register(
        self,
        address: int,
        value: Union[int, float, str],
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_req_write.WriteSingleRegisterResponse:
        """Write register (function code 0x06).

        :param address: Start address to read from
        :param value: Value to write
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_req_write.WriteSingleRegisterRequest(
            address, value, slave, **kwargs
        )
        return self.execute(request)

    def read_exception_status(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_other_msg.ReadExceptionStatusResponse:
        """Read Exception Status (function code 0x07).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_other_msg.ReadExceptionStatusRequest(slave, **kwargs)
        return self.execute(request)

    def diag_query_data(
        self, msg: bytearray, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnQueryDataResponse:
        """Diagnose query data (function code 0x08 - 0x00).

        :param msg: Message to be returned
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnQueryDataRequest(msg, slave, **kwargs)
        return self.execute(request)

    def diag_restart_communication(
        self, toggle: bool, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.RestartCommunicationsOptionResponse:
        """Diagnose restart communication (function code 0x08 - 0x01).

        :param toggle: True if toogled.
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.RestartCommunicationsOptionRequest(toggle, slave, **kwargs)
        return self.execute(request)

    def diag_read_diagnostic_register(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnDiagnosticRegisterResponse:
        """Diagnose read diagnostic register (function code 0x08 - 0x02).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnDiagnosticRegisterRequest(slave, **kwargs)
        return self.execute(request)

    def diag_change_ascii_input_delimeter(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ChangeAsciiInputDelimiterResponse:
        """Diagnose change ASCII input delimiter (function code 0x08 - 0x03).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ChangeAsciiInputDelimiterRequest(slave, **kwargs)
        return self.execute(request)

    def diag_force_listen_only(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ForceListenOnlyModeResponse:
        """Diagnose force listen only (function code 0x08 - 0x04).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ForceListenOnlyModeRequest(slave, **kwargs)
        return self.execute(request)

    def diag_clear_counters(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ClearCountersResponse:
        """Diagnose clear counters (function code 0x08 - 0x0A).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ClearCountersRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_bus_message_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnBusMessageCountResponse:
        """Diagnose read bus message count (function code 0x08 - 0x0B).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnBusMessageCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_bus_comm_error_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnBusCommunicationErrorCountResponse:
        """Diagnose read Bus Communication Error Count (function code 0x08 - 0x0C).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnBusCommunicationErrorCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_bus_exception_error_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnBusExceptionErrorCountResponse:
        """Diagnose read Bus Exception Error Count (function code 0x08 - 0x0D).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnBusExceptionErrorCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_slave_message_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnSlaveMessageCountResponse:
        """Diagnose read Slave Message Count (function code 0x08 - 0x0E).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnSlaveMessageCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_slave_no_response_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnSlaveNoReponseCountResponse:
        """Diagnose read Slave No Response Count (function code 0x08 - 0x0F).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnSlaveNoResponseCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_slave_nak_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnSlaveNAKCountResponse:
        """Diagnose read Slave NAK Count (function code 0x08 - 0x10).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnSlaveNAKCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_slave_busy_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnSlaveBusyCountResponse:
        """Diagnose read Slave Busy Count (function code 0x08 - 0x11).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnSlaveBusyCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_bus_char_overrun_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnSlaveBusCharacterOverrunCountResponse:
        """Diagnose read Bus Character Overrun Count (function code 0x08 - 0x12).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnSlaveBusCharacterOverrunCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_read_iop_overrun_count(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ReturnIopOverrunCountResponse:
        """Diagnose read Iop overrun count (function code 0x08 - 0x13).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ReturnIopOverrunCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_clear_overrun_counter(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.ClearOverrunCountResponse:
        """Diagnose Clear Overrun Counter and Flag (function code 0x08 - 0x14).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.ClearOverrunCountRequest(slave, **kwargs)
        return self.execute(request)

    def diag_getclear_modbus_response(
        self, slave: int = Defaults.Slave, **kwargs: any
    ) -> pdu_diag.GetClearModbusPlusResponse:
        """Diagnose Get/Clear modbus plus request (function code 0x08 - 0x15).

        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_diag.GetClearModbusPlusRequest(slave, **kwargs)
        return self.execute(request)

    # TBD missing functions
    # 0x0B Get Comm Event Counter (Serial Line only)
    # 0x0C Get Comm Event Log (Serial Line only)

    def write_coils(
        self,
        address: int,
        values: List[bool],
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_bit_write.WriteMultipleCoilsResponse:
        """Write coils (function code 0x0F).

        :param address: Start address to read from
        :param values: List of booleans to write
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_bit_write.WriteMultipleCoilsRequest(
            address, values, slave, **kwargs
        )
        return self.execute(request)

    def write_registers(
        self,
        address: int,
        values: List[Union[int, float, str]],
        slave: int = Defaults.Slave,
        **kwargs: any
    ) -> pdu_req_write.WriteMultipleRegistersResponse:
        """Write registers (function code 0x10).

        :param address: Start address to read from
        :param values: List of booleans to write
        :param slave: (optional) Modbus slave unit ID
        :param kwargs: (optional) Experimental parameters.
        :raises ModbusException:
        """
        if "unit" in kwargs:
            _logger.error("Please do not use unit=, convert to slave=.")
            slave = kwargs.pop("unit", slave)
        request = pdu_req_write.WriteMultipleRegistersRequest(
            address, values, slave, **kwargs
        )
        return self.execute(request)

    # Function codes descriptions
    # 0x11 Report Slave ID (Serial Line only)
    # 0x14 Read File Record
    # 0x15 Write File Record
    # 0x16 Mask Write Register
    # 0x17 Read/Write Multiple registers
    # 0x18 Read FIFO Queue
    # 0x2B Encapsulated Interface Transport
    # 0x2B / 0x0D CANopen General Reference Request and Response
    #     PDU
    # 0x2B / 0x0E Read Device Identification
    # MODBUS Exception Responses

    def readwrite_registers(
        self, *args, **kwargs
    ) -> pdu_reg_read.ReadWriteMultipleRegistersResponse:
        """Read/Write registers

        :param args:
        :param kwargs:
        :returns: A deferred response handle
        """
        request = pdu_reg_read.ReadWriteMultipleRegistersRequest(*args, **kwargs)
        return self.execute(request)

    def mask_write_register(
        self, *args, **kwargs
    ) -> pdu_req_write.MaskWriteRegisterResponse:
        """Mask write register.

        :args:
        :returns: A deferred response handle
        """
        request = pdu_req_write.MaskWriteRegisterRequest(*args, **kwargs)
        return self.execute(request)



"""Sync diag."""
import logging
import socket
import time

from pymodbus.client.tcp import ModbusTcpClient
from pymodbus.constants import Defaults
from pymodbus.exceptions import ConnectionException
from pymodbus.framer.socket_framer import ModbusSocketFramer


_logger = logging.getLogger(__name__)

LOG_MSGS = {
    "conn_msg": "Connecting to modbus device %s",
    "connfail_msg": "Connection to (%s, %s) failed: %s",
    "discon_msg": "Disconnecting from modbus device %s",
    "timelimit_read_msg": "Modbus device read took %.4f seconds, "
    "returned %s bytes in timelimit read",
    "timeout_msg": "Modbus device timeout after %.4f seconds, returned %s bytes %s",
    "delay_msg": "Modbus device read took %.4f seconds, "
    "returned %s bytes of %s expected",
    "read_msg": "Modbus device read took %.4f seconds, "
    "returned %s bytes of %s expected",
    "unexpected_dc_msg": "%s %s",
}


class ModbusTcpDiagClient(ModbusTcpClient):
    """Variant of pymodbus.client.ModbusTcpClient.

    With additional logging to diagnose network issues.

    The following events are logged:

    +---------+-----------------------------------------------------------------+
    | Level   | Events                                                          |
    +=========+=================================================================+
    | ERROR   | Failure to connect to modbus unit; unexpected disconnect by     |
    |         | modbus unit                                                     |
    +---------+-----------------------------------------------------------------+
    | WARNING | Timeout on normal read; read took longer than warn_delay_limit  |
    +---------+-----------------------------------------------------------------+
    | INFO    | Connection attempt to modbus unit; disconnection from modbus    |
    |         | unit; each time limited read                                    |
    +---------+-----------------------------------------------------------------+
    | DEBUG   | Normal read with timing information                             |
    +---------+-----------------------------------------------------------------+

    Reads are differentiated between "normal", which reads a specified number of
    bytes, and "time limited", which reads all data for a duration equal to the
    timeout period configured for this instance.
    """

    def __init__(
        self,
        host="127.0.0.1",
        port=Defaults.TcpPort,
        framer=ModbusSocketFramer,
        **kwargs,
    ):
        """Initialize a client instance.

        The keys of LOG_MSGS can be used in kwargs to customize the messages.

        :param host: The host to connect to (default 127.0.0.1)
        :param port: The modbus port to connect to (default 502)
        :param source_address: The source address tuple to bind to (default ("", 0))
        :param timeout: The timeout to use for this socket (default Defaults.Timeout)
        :param warn_delay_limit: Log reads that take longer than this as warning.
               Default True sets it to half of "timeout". None never logs these as
               warning, 0 logs everything as warning.
        :param framer: The modbus framer to use (default ModbusSocketFramer)

        .. note:: The host argument will accept ipv4 and ipv6 hosts
        """
        self.warn_delay_limit = kwargs.get("warn_delay_limit", True)
        super().__init__(host, port, framer, **kwargs)
        if self.warn_delay_limit is True:
            self.warn_delay_limit = self.params.timeout / 2

        # Set logging messages, defaulting to LOG_MSGS
        for (k_item, v_item) in LOG_MSGS.items():
            self.__dict__[k_item] = kwargs.get(k_item, v_item)

    def connect(self):
        """Connect to the modbus tcp server.

        :returns: True if connection succeeded, False otherwise
        """
        if self.socket:
            return True
        try:
            _logger.info(LOG_MSGS["conn_msg"], self)
            self.socket = socket.create_connection(
                (self.params.host, self.params.port),
                timeout=self.params.timeout,
                source_address=self.params.source_address,
            )
        except socket.error as msg:
            _logger.error(
                LOG_MSGS["connfail_msg"], self.params.host, self.params.port, msg
            )
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection."""
        if self.socket:
            _logger.info(LOG_MSGS["discon_msg"], self)
            self.socket.close()
        self.socket = None

    def recv(self, size):
        """Receive data."""
        try:
            start = time.time()

            result = super().recv(size)

            delay = time.time() - start
            if self.warn_delay_limit is not None and delay >= self.warn_delay_limit:
                self._log_delayed_response(len(result), size, delay)
            elif not size:
                _logger.debug(LOG_MSGS["timelimit_read_msg"], delay, len(result))
            else:
                _logger.debug(LOG_MSGS["read_msg"], delay, len(result), size)

            return result
        except ConnectionException as exc:
            # Only log actual network errors, "if not self.socket" then it"s a internal code issue
            if "Connection unexpectedly closed" in exc.string:
                _logger.error(LOG_MSGS["unexpected_dc_msg"], self, exc)
            raise ConnectionException from exc

    def _log_delayed_response(self, result_len, size, delay):
        """Log delayed response."""
        if not size and result_len > 0:
            _logger.info(LOG_MSGS["timelimit_read_msg"], delay, result_len)
        elif (
            (not result_len) or (size and result_len < size)
        ) and delay >= self.params.timeout:
            size_txt = size if size else "in timelimit read"
            read_type = f"of {size_txt} expected"
            _logger.warning(LOG_MSGS["timeout_msg"], delay, result_len, read_type)
        else:
            _logger.warning(LOG_MSGS["delay_msg"], delay, result_len, size)

    def __str__(self):
        """Build a string representation of the connection.

        :returns: The string representation
        """
        return f"ModbusTcpDiagClient({self.params.host}:{self.params.port})"


def get_client():
    """Return an appropriate client based on logging level.

    This will be ModbusTcpDiagClient by default, or the parent class
    if the log level is such that the diagnostic client will not log
    anything.

    :returns: ModbusTcpClient or a child class thereof
    """
    return (
        ModbusTcpDiagClient if _logger.isEnabledFor(logging.ERROR) else ModbusTcpClient
    )


# --------------------------------------------------------------------------- #
# Exported symbols
# --------------------------------------------------------------------------- #

__all__ = ["ModbusTcpDiagClient", "get_client"]



#!/usr/bin/env python3
"""Pymodbus Asynchronous Server Example.

The asynchronous server is a high performance implementation using the
twisted library as its backend.  This allows it to scale to many thousands
of nodes which can be helpful for testing monitoring software.
"""
# --------------------------------------------------------------------------- #
# import the various server implementations
# --------------------------------------------------------------------------- #
import logging

from custom_message import CustomModbusRequest

from pymodbus.datastore import (
    ModbusSequentialDataBlock,
    ModbusServerContext,
    ModbusSlaveContext,
)

# from pymodbus.server.asynchronous import StartUdpServer
# from pymodbus.server.asynchronous import StartSerialServer
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.server.asynchronous import StartTcpServer
from pymodbus.version import version


# from pymodbus.transaction import (
#    ModbusRtuFramer,
#    ModbusAsciiFramer,
#    ModbusBinaryFramer,
# )

# --------------------------------------------------------------------------- #
# configure the service logging
# --------------------------------------------------------------------------- #
FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.DEBUG)


def run_async_server():
    """Run async server."""
    # ----------------------------------------------------------------------- #
    # initialize your data store
    # ----------------------------------------------------------------------- #
    # The datastores only respond to the addresses that they are initialized to
    # Therefore, if you initialize a DataBlock to addresses from 0x00 to 0xFF,
    # a request to 0x100 will respond with an invalid address exception.
    # This is because many devices exhibit this kind of behavior (but not all)
    #
    #     block = ModbusSequentialDataBlock(0x00, [0]*0xff)
    #
    # Continuing, you can choose to use a sequential or a sparse DataBlock in
    # your data context.  The difference is that the sequential has no gaps in
    # the data while the sparse can. Once again, there are devices that exhibit
    # both forms of behavior::
    #
    #     block = ModbusSparseDataBlock({0x00: 0, 0x05: 1})
    #     block = ModbusSequentialDataBlock(0x00, [0]*5)
    #
    # Alternately, you can use the factory methods to initialize the DataBlocks
    # or simply do not pass them to have them initialized to 0x00 on the full
    # address range::
    #
    #     store = ModbusSlaveContext(di = ModbusSequentialDataBlock.create())
    #     store = ModbusSlaveContext()
    #
    # Finally, you are allowed to use the same DataBlock reference for every
    # table or you you may use a separate DataBlock for each table.
    # This depends if you would like functions to be able to access and modify
    # the same data or not::
    #
    #     block = ModbusSequentialDataBlock(0x00, [0]*0xff)
    #     store = ModbusSlaveContext(di=block, co=block, hr=block, ir=block)
    #
    # The server then makes use of a server context that allows the server to
    # respond with different slave contexts for different unit ids. By default
    # it will return the same context for every unit id supplied (broadcast
    # mode).
    # However, this can be overloaded by setting the single flag to False
    # and then supplying a dictionary of unit id to context mapping::
    #
    #     slaves  = {
    #         0x01: ModbusSlaveContext(...),
    #         0x02: ModbusSlaveContext(...),
    #         0x03: ModbusSlaveContext(...),
    #     }
    #     context = ModbusServerContext(slaves=slaves, single=False)
    #
    # The slave context can also be initialized in zero_mode which means that a
    # request to address(0-7) will map to the address (0-7). The default is
    # False which is based on section 4.4 of the specification, so address(0-7)
    # will map to (1-8)::
    #
    #     store = ModbusSlaveContext(..., zero_mode=True)
    # ----------------------------------------------------------------------- #
    store = ModbusSlaveContext(
        co=ModbusSequentialDataBlock(0, [17] * 100),
        di=ModbusSequentialDataBlock(0, [17] * 100),
        hr=ModbusSequentialDataBlock(0, [17] * 100),
        ir=ModbusSequentialDataBlock(0, [17] * 100),
    )
    store.register(
        CustomModbusRequest.function_code,
        "cm",
        ModbusSequentialDataBlock(0, [17] * 100),
    )
    context = ModbusServerContext(slaves=store, single=True)

    # ----------------------------------------------------------------------- #
    # initialize the server information
    # ----------------------------------------------------------------------- #
    # If you don"t set this or any fields, they are defaulted to empty strings.
    # ----------------------------------------------------------------------- #
    identity = ModbusDeviceIdentification(
        info_name={
            "VendorName": "Pymodbus",
            "ModelName": "Pymodbus Server",
            "MajorMinorRevision": version.short(),
            "ProductCode": "PM",
            "VendorUrl": "https://github.com/riptideio/pymodbus/",
            "ProductName": "Pymodbus Server",
        }
    )

    # ----------------------------------------------------------------------- #
    # run the server you want
    # ----------------------------------------------------------------------- #

    # TCP Server

    StartTcpServer(
        context,
        identity=identity,
        address=("localhost", 5020),
        custom_functions=[CustomModbusRequest],
    )

    # TCP Server with deferred reactor run

    # from twisted.internet import reactor
    # StartTcpServer(context, identity=identity, address=("localhost", 5020),
    #                defer_reactor_run=True)
    # reactor.run()

    # Server with RTU framer
    # StartTcpServer(context, identity=identity, address=("localhost", 5020),
    #                framer=ModbusRtuFramer)

    # UDP Server
    # StartUdpServer(context, identity=identity, address=("127.0.0.1", 5020))

    # RTU Server
    # StartSerialServer(context, identity=identity,
    #                   port="/dev/ttyp0", framer=ModbusRtuFramer)

    # ASCII Server
    # StartSerialServer(context, identity=identity,
    #                   port="/dev/ttyp0", framer=ModbusAsciiFramer)

    # Binary Server
    # StartSerialServer(context, identity=identity,
    #                   port="/dev/ttyp0", framer=ModbusBinaryFramer)


if __name__ == "__main__":
    run_async_server()



"""Collection of transaction based abstractions."""
# pylint: disable=missing-type-doc
from functools import partial
import logging
import socket
import struct
from threading import RLock
import time

from pymodbus.constants import Defaults
from pymodbus.exceptions import (
    InvalidMessageReceivedException,
    ModbusIOException,
    NotImplementedException,
)
from pymodbus.framer.ascii_framer import ModbusAsciiFramer
from pymodbus.framer.binary_framer import ModbusBinaryFramer
from pymodbus.framer.rtu_framer import ModbusRtuFramer
from pymodbus.framer.socket_framer import ModbusSocketFramer
from pymodbus.framer.tls_framer import ModbusTlsFramer
from pymodbus.utilities import ModbusTransactionState, hexlify_packets


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------- #
# The Global Transaction Manager
# --------------------------------------------------------------------------- #
class ModbusTransactionManager:
    """Implement a transaction for a manager.

    The transaction protocol can be represented by the following pseudo code::

        count = 0
        do
          result = send(message)
          if (timeout or result == bad)
             count++
          else break
        while (count < 3)

    This module helps to abstract this away from the framer and protocol.
    """

    def __init__(self, client, **kwargs):
        """Initialize an instance of the ModbusTransactionManager.

        :param client: The client socket wrapper
        :param retry_on_empty: Should the client retry on empty
        :param retries: The number of retries to allow
        """
        self.tid = Defaults.TransactionId
        self.client = client
        self.backoff = kwargs.get("backoff", Defaults.Backoff) or 0.3
        self.retry_on_empty = kwargs.get("retry_on_empty", Defaults.RetryOnEmpty)
        self.retry_on_invalid = kwargs.get("retry_on_invalid", Defaults.RetryOnInvalid)
        self.retries = kwargs.get("retries", Defaults.Retries) or 1
        self.reset_socket = kwargs.get("reset_socket", True)
        self._transaction_lock = RLock()
        self._no_response_devices = []
        if client:
            self._set_adu_size()

    def _set_adu_size(self):
        """Set adu size."""
        # base ADU size of modbus frame in bytes
        if isinstance(self.client.framer, ModbusSocketFramer):
            self.base_adu_size = 7  # tid(2), pid(2), length(2), uid(1)
        elif isinstance(self.client.framer, ModbusRtuFramer):
            self.base_adu_size = 3  # address(1), CRC(2)
        elif isinstance(self.client.framer, ModbusAsciiFramer):
            self.base_adu_size = 7  # start(1)+ Address(2), LRC(2) + end(2)
        elif isinstance(self.client.framer, ModbusBinaryFramer):
            self.base_adu_size = 5  # start(1) + Address(1), CRC(2) + end(1)
        elif isinstance(self.client.framer, ModbusTlsFramer):
            self.base_adu_size = 0  # no header and footer
        else:
            self.base_adu_size = -1

    def _calculate_response_length(self, expected_pdu_size):
        """Calculate response length."""
        if self.base_adu_size == -1:
            return None
        return self.base_adu_size + expected_pdu_size

    def _calculate_exception_length(self):
        """Return the length of the Modbus Exception Response according to the type of Framer."""
        if isinstance(self.client.framer, (ModbusSocketFramer, ModbusTlsFramer)):
            return self.base_adu_size + 2  # Fcode(1), ExceptionCode(1)
        if isinstance(self.client.framer, ModbusAsciiFramer):
            return self.base_adu_size + 4  # Fcode(2), ExceptionCode(2)
        if isinstance(self.client.framer, (ModbusRtuFramer, ModbusBinaryFramer)):
            return self.base_adu_size + 2  # Fcode(1), ExceptionCode(1)
        return None

    def _validate_response(self, request, response, exp_resp_len):
        """Validate Incoming response against request.

        :param request: Request sent
        :param response: Response received
        :param exp_resp_len: Expected response length
        :return: New transactions state
        """
        if not response:
            return False

        mbap = self.client.framer.decode_data(response)
        if (
            mbap.get("unit") != request.unit_id
            or mbap.get("fcode") & 0x7F != request.function_code
        ):
            return False

        if "length" in mbap and exp_resp_len:
            return mbap.get("length") == exp_resp_len
        return True

    def execute(self, request):  # pylint: disable=too-complex
        """Start the producer to send the next request to consumer.write(Frame(request))."""
        with self._transaction_lock:
            try:
                txt = (
                    "Current transaction state - "
                    f"{ModbusTransactionState.to_string(self.client.state)}"
                )
                _logger.debug(txt)
                retries = self.retries
                request.transaction_id = self.getNextTID()
                txt = f"Running transaction {request.transaction_id}"
                _logger.debug(txt)
                if _buffer := hexlify_packets(
                    self.client.framer._buffer  # pylint: disable=protected-access
                ):
                    txt = f"Clearing current Frame: - {_buffer}"
                    _logger.debug(txt)
                    self.client.framer.resetFrame()
                if broadcast := (
                    self.client.params.broadcast_enable and not request.unit_id
                ):
                    self._transact(request, None, broadcast=True)
                    response = b"Broadcast write sent - no response expected"
                else:
                    expected_response_length = None
                    if not isinstance(self.client.framer, ModbusSocketFramer):
                        if hasattr(request, "get_response_pdu_size"):
                            response_pdu_size = request.get_response_pdu_size()
                            if isinstance(self.client.framer, ModbusAsciiFramer):
                                response_pdu_size = response_pdu_size * 2
                            if response_pdu_size:
                                expected_response_length = (
                                    self._calculate_response_length(response_pdu_size)
                                )
                    if (  # pylint: disable=simplifiable-if-statement
                        request.unit_id in self._no_response_devices
                    ):
                        full = True
                    else:
                        full = False
                    c_str = str(self.client)
                    if "modbusudpclient" in c_str.lower().strip():
                        full = True
                        if not expected_response_length:
                            expected_response_length = Defaults.ReadSize
                    response, last_exception = self._transact(
                        request,
                        expected_response_length,
                        full=full,
                        broadcast=broadcast,
                    )
                    while retries > 0:
                        valid_response = self._validate_response(
                            request, response, expected_response_length
                        )
                        if valid_response:
                            if (
                                request.unit_id in self._no_response_devices
                                and response
                            ):
                                self._no_response_devices.remove(request.unit_id)
                                _logger.debug("Got response!!!")
                            break
                        if not response:
                            if request.unit_id not in self._no_response_devices:
                                self._no_response_devices.append(request.unit_id)
                            if self.retry_on_empty:
                                response, last_exception = self._retry_transaction(
                                    retries,
                                    "empty",
                                    request,
                                    expected_response_length,
                                    full=full,
                                )
                                retries -= 1
                            else:
                                # No response received and retries not enabled
                                break
                        elif self.retry_on_invalid:
                            response, last_exception = self._retry_transaction(
                                retries,
                                "invalid",
                                request,
                                expected_response_length,
                                full=full,
                            )
                            retries -= 1
                        else:
                            break
                        # full = False
                    addTransaction = partial(  # pylint: disable=invalid-name
                        self.addTransaction,
                        tid=request.transaction_id,
                    )
                    self.client.framer.processIncomingPacket(
                        response, addTransaction, request.unit_id
                    )
                    if not (response := self.getTransaction(request.transaction_id)):
                        if len(self.transactions):
                            response = self.getTransaction(tid=0)
                        else:
                            last_exception = last_exception or (
                                "No Response received from the remote unit"
                                "/Unable to decode response"
                            )
                            response = ModbusIOException(
                                last_exception, request.function_code
                            )
                        if self.reset_socket:
                            self.client.close()
                    if hasattr(self.client, "state"):
                        _logger.debug(
                            "Changing transaction state from "
                            '"PROCESSING REPLY" to '
                            '"TRANSACTION_COMPLETE"'
                        )
                        self.client.state = ModbusTransactionState.TRANSACTION_COMPLETE

                return response
            except ModbusIOException as exc:
                # Handle decode errors in processIncomingPacket method
                _logger.exception(exc)
                self.client.state = ModbusTransactionState.TRANSACTION_COMPLETE
                if self.reset_socket:
                    self.client.close()
                return exc

    def _retry_transaction(self, retries, reason, packet, response_length, full=False):
        """Retry transaction."""
        txt = f"Retry on {reason} response - {retries}"
        _logger.debug(txt)
        _logger.debug(
            'Changing transaction state from "WAITING_FOR_REPLY" to "RETRYING"'
        )
        self.client.state = ModbusTransactionState.RETRYING
        if self.backoff:
            delay = 2 ** (self.retries - retries) * self.backoff
            time.sleep(delay)
            txt = f"Sleeping {delay}"
            _logger.debug(txt)
        self.client.connect()
        if hasattr(self.client, "_in_waiting"):
            if (
                in_waiting := self.client._in_waiting()  # pylint: disable=protected-access
            ):
                if response_length == in_waiting:
                    result = self._recv(response_length, full)
                    return result, None
        return self._transact(packet, response_length, full=full)

    def _transact(
        self, packet, response_length, full=False, broadcast=False
    ):  # pylint: disable=too-complex
        """Do a Write and Read transaction.

        :param packet: packet to be sent
        :param response_length:  Expected response length
        :param full: the target device was notorious for its no response. Dont
            waste time this time by partial querying
        :param broadcast:
        :return: response
        """
        last_exception = None
        try:
            self.client.connect()
            packet = self.client.framer.buildPacket(packet)
            if _logger.isEnabledFor(logging.DEBUG):
                txt = f"SEND: {hexlify_packets(packet)}"
                _logger.debug(txt)
            size = self._send(packet)
            if (
                isinstance(size, bytes)
                and self.client.state == ModbusTransactionState.RETRYING
            ):
                _logger.debug(
                    "Changing transaction state from "
                    '"RETRYING" to "PROCESSING REPLY"'
                )
                self.client.state = ModbusTransactionState.PROCESSING_REPLY
                return size, None
            if broadcast:
                if size:
                    _logger.debug(
                        'Changing transaction state from "SENDING" '
                        'to "TRANSACTION_COMPLETE"'
                    )
                    self.client.state = ModbusTransactionState.TRANSACTION_COMPLETE
                return b"", None
            if size:
                _logger.debug(
                    'Changing transaction state from "SENDING" '
                    'to "WAITING FOR REPLY"'
                )
                self.client.state = ModbusTransactionState.WAITING_FOR_REPLY
            if (
                hasattr(self.client, "handle_local_echo")
                and self.client.handle_local_echo is True
            ):
                if self._recv(size, full) != packet:
                    return b"", "Wrong local echo"
            result = self._recv(response_length, full)
            # result2 = self._recv(response_length, full)
            if _logger.isEnabledFor(logging.DEBUG):
                txt = f"RECV: {hexlify_packets(result)}"
                _logger.debug(txt)

        except (
            socket.error,
            ModbusIOException,
            InvalidMessageReceivedException,
        ) as msg:
            if self.reset_socket:
                self.client.close()
            txt = f"Transaction failed. ({msg}) "
            _logger.debug(txt)
            last_exception = msg
            result = b""
        return result, last_exception

    def _send(self, packet, retrying=False):  # pylint: disable=unused-argument
        """Send."""
        return self.client.framer.sendPacket(packet)

    def _recv(self, expected_response_length, full):  # pylint: disable=too-complex
        """Receive."""
        total = None
        if not full:
            exception_length = self._calculate_exception_length()
            if isinstance(self.client.framer, ModbusSocketFramer):
                min_size = 8
            elif isinstance(self.client.framer, ModbusRtuFramer):
                min_size = 4
            elif isinstance(self.client.framer, ModbusAsciiFramer):
                min_size = 5
            elif isinstance(self.client.framer, ModbusBinaryFramer):
                min_size = 3
            else:
                min_size = expected_response_length

            read_min = self.client.framer.recvPacket(min_size)
            if len(read_min) != min_size:
                msg_start = "Incomplete message" if read_min else "No response"
                raise InvalidMessageReceivedException(
                    f"{msg_start} received, expected at least {min_size} bytes "
                    f"({len(read_min)} received)"
                )
            if read_min:
                if isinstance(self.client.framer, ModbusSocketFramer):
                    func_code = int(read_min[-1])
                elif isinstance(self.client.framer, ModbusRtuFramer):
                    func_code = int(read_min[1])
                elif isinstance(self.client.framer, ModbusAsciiFramer):
                    func_code = int(read_min[3:5], 16)
                elif isinstance(self.client.framer, ModbusBinaryFramer):
                    func_code = int(read_min[-1])
                else:
                    func_code = -1

                if func_code < 0x80:  # Not an error
                    if isinstance(self.client.framer, ModbusSocketFramer):
                        # Omit UID, which is included in header size
                        h_size = (
                            self.client.framer._hsize  # pylint: disable=protected-access
                        )
                        length = struct.unpack(">H", read_min[4:6])[0] - 1
                        expected_response_length = h_size + length
                    elif expected_response_length is None and isinstance(
                        self.client.framer, ModbusRtuFramer
                    ):
                        try:
                            expected_response_length = (
                                self.client.framer.get_expected_response_length(
                                    read_min
                                )
                            )
                        except IndexError:
                            # Could not determine response length with available bytes
                            pass
                    if expected_response_length is not None:
                        expected_response_length -= min_size
                        total = expected_response_length + min_size
                else:
                    expected_response_length = exception_length - min_size
                    total = expected_response_length + min_size
            else:
                total = expected_response_length
        else:
            read_min = b""
            total = expected_response_length
        result = self.client.framer.recvPacket(expected_response_length)
        result = read_min + result
        actual = len(result)
        if total is not None and actual != total:
            msg_start = "Incomplete message" if actual else "No response"
            txt = (
                "{msg_start} received, "
                f"Expected {total} bytes Received "
                f"{actual} bytes !!!!"
            )
            _logger.debug(txt)
        elif not actual:
            # If actual == 0 and total is not None then the above
            # should be triggered, so total must be None here
            _logger.debug("No response received to unbounded read !!!!")
        if self.client.state != ModbusTransactionState.PROCESSING_REPLY:
            _logger.debug(
                "Changing transaction state from "
                '"WAITING FOR REPLY" to "PROCESSING REPLY"'
            )
            self.client.state = ModbusTransactionState.PROCESSING_REPLY
        return result

    def addTransaction(self, request, tid=None):  # pylint: disable=invalid-name
        """Add a transaction to the handler.

        This holds the request in case it needs to be resent.
        After being sent, the request is removed.

        :param request: The request to hold on to
        :param tid: The overloaded transaction id to use
        :raises NotImplementedException:
        """
        raise NotImplementedException("addTransaction")

    def getTransaction(self, tid):  # pylint: disable=invalid-name
        """Return a transaction matching the referenced tid.

        If the transaction does not exist, None is returned

        :param tid: The transaction to retrieve
        :raises NotImplementedException:
        """
        raise NotImplementedException("getTransaction")

    def delTransaction(self, tid):  # pylint: disable=invalid-name
        """Remove a transaction matching the referenced tid.

        :param tid: The transaction to remove
        :raises NotImplementedException:
        """
        raise NotImplementedException("delTransaction")

    def getNextTID(self):  # pylint: disable=invalid-name
        """Retrieve the next unique transaction identifier.

        This handles incrementing the identifier after
        retrieval

        :returns: The next unique transaction identifier
        """
        self.tid = (self.tid + 1) & 0xFFFF
        return self.tid

    def reset(self):
        """Reset the transaction identifier."""
        self.tid = Defaults.TransactionId
        self.transactions = type(  # pylint: disable=attribute-defined-outside-init
            self.transactions
        )()


class DictTransactionManager(ModbusTransactionManager):
    """Implements a transaction for a manager.

    Where the results are keyed based on the supplied transaction id.
    """

    def __init__(self, client, **kwargs):
        """Initialize an instance of the ModbusTransactionManager.

        :param client: The client socket wrapper
        """
        self.transactions = {}
        super().__init__(client, **kwargs)

    def __iter__(self):
        """Iterate over the current managed transactions.

        :returns: An iterator of the managed transactions
        """
        return iter(self.transactions.keys())

    def addTransaction(self, request, tid=None):
        """Add a transaction to the handler.

        This holds the requets in case it needs to be resent.
        After being sent, the request is removed.

        :param request: The request to hold on to
        :param tid: The overloaded transaction id to use
        """
        tid = tid if tid is not None else request.transaction_id
        txt = f"Adding transaction {tid}"
        _logger.debug(txt)
        self.transactions[tid] = request

    def getTransaction(self, tid):
        """Return a transaction matching the referenced tid.

        If the transaction does not exist, None is returned

        :param tid: The transaction to retrieve

        """
        txt = f"Getting transaction {tid}"
        _logger.debug(txt)
        if not tid:
            if self.transactions:
                return self.transactions.popitem()[1]
            return None
        return self.transactions.pop(tid, None)

    def delTransaction(self, tid):
        """Remove a transaction matching the referenced tid.

        :param tid: The transaction to remove
        """
        txt = f"deleting transaction {tid}"
        _logger.debug(txt)

        self.transactions.pop(tid, None)


class FifoTransactionManager(ModbusTransactionManager):
    """Implements a transaction.

    For a manager where the results are returned in a FIFO manner.
    """

    def __init__(self, client, **kwargs):
        """Initialize an instance of the ModbusTransactionManager.

        :param client: The client socket wrapper
        """
        super().__init__(client, **kwargs)
        self.transactions = []

    def __iter__(self):
        """Iterate over the current managed transactions.

        :returns: An iterator of the managed transactions
        """
        return iter(self.transactions)

    def addTransaction(self, request, tid=None):
        """Add a transaction to the handler.

        This holds the requets in case it needs to be resent.
        After being sent, the request is removed.

        :param request: The request to hold on to
        :param tid: The overloaded transaction id to use
        """
        tid = tid if tid is not None else request.transaction_id
        txt = f"Adding transaction {tid}"
        _logger.debug(txt)

        self.transactions.append(request)

    def getTransaction(self, tid):
        """Return a transaction matching the referenced tid.

        If the transaction does not exist, None is returned

        :param tid: The transaction to retrieve
        """
        return self.transactions.pop(0) if self.transactions else None

    def delTransaction(self, tid):
        """Remove a transaction matching the referenced tid.

        :param tid: The transaction to remove
        """
        txt = f"Deleting transaction {tid}"
        _logger.debug(txt)
        if self.transactions:
            self.transactions.pop(0)


# --------------------------------------------------------------------------- #
# Exported symbols
# --------------------------------------------------------------------------- #


__all__ = [
    "FifoTransactionManager",
    "DictTransactionManager",
    "ModbusSocketFramer",
    "ModbusTlsFramer",
    "ModbusRtuFramer",
    "ModbusAsciiFramer",
    "ModbusBinaryFramer",
]



"""Modbus client async serial communication."""
from functools import partial
import asyncio
import time
import logging

from serial_asyncio import create_serial_connection
import serial

from pymodbus.client.base import ModbusBaseClient, ModbusClientProtocol
from pymodbus.constants import Defaults
from pymodbus.framer import ModbusFramer
from pymodbus.framer.rtu_framer import ModbusRtuFramer
from pymodbus.exceptions import ConnectionException
from pymodbus.utilities import ModbusTransactionState, hexlify_packets


_logger = logging.getLogger(__name__)


class AsyncModbusSerialClient(ModbusBaseClient):
    """**AsyncModbusSerialClient**.

    :param port: Serial port used for communication.
    :param framer: (optional) Framer class.
    :param baudrate: (optional) Bits pr second.
    :param bytesize: (optional) Number of bits pr byte 7-8.
    :param parity: (optional) 'E'ven, 'O'dd or 'N'one
    :param stopbits: (optional) Number of stop bits 0-2¡.
    :param handle_local_echo: (optional) Discard local echo from dongle.
    :param kwargs: (optional) Experimental parameters

    The serial communication is RS-485 based, and usually used vith a usb RS485 dongle.

    Example::

        from pymodbus.client import AsyncModbusSerialClient

        async def run():
            client = AsyncModbusSerialClient("dev/serial0")

            await client.connect()
            ...
            await client.close()
    """

    transport = None
    framer = None

    def __init__(
        self,
        port: str,
        framer: ModbusFramer = ModbusRtuFramer,
        baudrate: int = Defaults.Baudrate,
        bytesize: int = Defaults.Bytesize,
        parity: chr = Defaults.Parity,
        stopbits: int = Defaults.Stopbits,
        handle_local_echo: bool = Defaults.HandleLocalEcho,
        **kwargs: any,
    ) -> None:
        """Initialize Asyncio Modbus Serial Client."""
        self.protocol = None
        super().__init__(framer=framer, **kwargs)
        self.params.port = port
        self.params.baudrate = baudrate
        self.params.bytesize = bytesize
        self.params.parity = parity
        self.params.stopbits = stopbits
        self.params.handle_local_echo = handle_local_echo
        self.loop = None
        self._connected_event = asyncio.Event()

    async def close(self):  # pylint: disable=invalid-overridden-method
        """Stop connection."""
        if self.connected and self.protocol and self.protocol.transport:
            self.protocol.transport.close()

    def _create_protocol(self):
        """Create protocol."""
        protocol = ModbusClientProtocol(framer=self.params.framer)
        protocol.factory = self
        return protocol

    @property
    def connected(self):
        """Connect internal."""
        return self._connected_event.is_set()

    async def connect(self):  # pylint: disable=invalid-overridden-method
        """Connect Async client."""
        # get current loop, if there are no loop a RuntimeError will be raised
        self.loop = asyncio.get_running_loop()

        _logger.debug("Starting serial connection")
        try:
            await create_serial_connection(
                self.loop,
                self._create_protocol,
                self.params.port,
                baudrate=self.params.baudrate,
                bytesize=self.params.bytesize,
                stopbits=self.params.stopbits,
                parity=self.params.parity,
                **self.params.kwargs,
            )
            await self._connected_event.wait()
            txt = f"Connected to {self.params.port}"
            _logger.info(txt)
        except Exception as exc:  # pylint: disable=broad-except
            txt = f"Failed to connect: {exc}"
            _logger.warning(txt)

    def protocol_made_connection(self, protocol):
        """Notify successful connection."""
        _logger.info("Serial connected.")
        if not self.connected:
            self._connected_event.set()
            self.protocol = protocol
        else:
            _logger.error("Factory protocol connect callback called while connected.")

    def protocol_lost_connection(self, protocol):
        """Notify lost connection."""
        if self.connected:
            _logger.info("Serial lost connection.")
            if protocol is not self.protocol:
                _logger.error("Serial: protocol is not self.protocol.")

            self._connected_event.clear()
            self.protocol = None
            # if self.host:
            #     asyncio.asynchronous(self._reconnect())
        else:
            _logger.error("Serial, lost_connection but not connected.")


class ModbusSerialClient(ModbusBaseClient):
    """**ModbusSerialClient**.

    :param port: Serial port used for communication.
    :param framer: (optional) Framer class.
    :param baudrate: (optional) Bits pr second.
    :param bytesize: (optional) Number of bits pr byte 7-8.
    :param parity: (optional) 'E'ven, 'O'dd or 'N'one
    :param stopbits: (optional) Number of stop bits 0-2¡.
    :param handle_local_echo: (optional) Discard local echo from dongle.
    :param kwargs: (optional) Experimental parameters

    The serial communication is RS-485 based, and usually used vith a usb RS485 dongle.

    Example::

        from pymodbus.client import ModbusSerialClient

        def run():
            client = ModbusSerialClient("dev/serial0")

            client.connect()
            ...
            client.close()
    """

    state = ModbusTransactionState.IDLE
    inter_char_timeout = 0
    silent_interval = 0

    def __init__(
        self,
        port: str,
        framer: ModbusFramer = ModbusRtuFramer,
        baudrate: int = Defaults.Baudrate,
        bytesize: int = Defaults.Bytesize,
        parity: chr = Defaults.Parity,
        stopbits: int = Defaults.Stopbits,
        handle_local_echo: bool = Defaults.HandleLocalEcho,
        **kwargs: any,
    ) -> None:
        """Initialize Modbus Serial Client."""
        super().__init__(framer=framer, **kwargs)
        self.params.port = port
        self.params.baudrate = baudrate
        self.params.bytesize = bytesize
        self.params.parity = parity
        self.params.stopbits = stopbits
        self.params.handle_local_echo = handle_local_echo
        self.socket = None

        self.last_frame_end = None
        if isinstance(self.framer, ModbusRtuFramer):
            if self.params.baudrate > 19200:
                self.silent_interval = 1.75 / 1000  # ms
            else:
                self._t0 = float((1 + 8 + 2)) / self.params.baudrate
                self.inter_char_timeout = 1.5 * self._t0
                self.silent_interval = 3.5 * self._t0
            self.silent_interval = round(self.silent_interval, 6)

    @property
    def connected(self):
        """Connect internal."""
        return self.connect()

    def connect(self):
        """Connect to the modbus serial server."""
        if self.socket:
            return True
        try:
            self.socket = serial.serial_for_url(
                self.params.port,
                timeout=self.params.timeout,
                bytesize=self.params.bytesize,
                stopbits=self.params.stopbits,
                baudrate=self.params.baudrate,
                parity=self.params.parity,
            )
            if isinstance(self.framer, ModbusRtuFramer):
                if self.params.strict:
                    self.socket.interCharTimeout = self.inter_char_timeout
                self.last_frame_end = None
        except serial.SerialException as msg:
            _logger.error(msg)
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection."""
        if self.socket:
            self.socket.close()
        self.socket = None

    def _in_waiting(self):
        """Return _in_waiting."""
        in_waiting = "in_waiting" if hasattr(self.socket, "in_waiting") else "inWaiting"

        if in_waiting == "in_waiting":
            waitingbytes = getattr(self.socket, in_waiting)
        else:
            waitingbytes = getattr(self.socket, in_waiting)()
        return waitingbytes

    def send(self, request):
        """Send data on the underlying socket.

        If receive buffer still holds some data then flush it.

        Sleep if last send finished less than 3.5 character times ago.
        """
        super().send(request)
        if not self.socket:
            raise ConnectionException(str(self))
        if request:
            try:
                if waitingbytes := self._in_waiting():
                    result = self.socket.read(waitingbytes)
                    if self.state == ModbusTransactionState.RETRYING:
                        txt = f"Sending available data in recv buffer {hexlify_packets(result)}"
                        _logger.debug(txt)
                        return result
                    if _logger.isEnabledFor(logging.WARNING):
                        txt = f"Cleanup recv buffer before send: {hexlify_packets(result)}"
                        _logger.warning(txt)
            except NotImplementedError:
                pass
            if self.state != ModbusTransactionState.SENDING:
                _logger.debug('New Transaction state "SENDING"')
                self.state = ModbusTransactionState.SENDING
            size = self.socket.write(request)
            return size
        return 0

    def _wait_for_data(self):
        """Wait for data."""
        size = 0
        more_data = False
        if self.params.timeout is not None and self.params.timeout:
            condition = partial(
                lambda start, timeout: (time.time() - start) <= timeout,
                timeout=self.params.timeout,
            )
        else:
            condition = partial(lambda dummy1, dummy2: True, dummy2=None)
        start = time.time()
        while condition(start):
            available = self._in_waiting()
            if (more_data and not available) or (more_data and available == size):
                break
            if available and available != size:
                more_data = True
                size = available
            time.sleep(0.01)
        return size

    def recv(self, size):
        """Read data from the underlying descriptor."""
        super().recv(size)
        if not self.socket:
            raise ConnectionException(
                self.__str__()  # pylint: disable=unnecessary-dunder-call
            )
        if size is None:
            size = self._wait_for_data()
        result = self.socket.read(size)
        return result

    def is_socket_open(self):
        """Check if socket is open."""
        if self.socket:
            if hasattr(self.socket, "is_open"):
                return self.socket.is_open
            return self.socket.isOpen()
        return False

    def __str__(self):
        """Build a string representation of the connection."""
        return f"ModbusSerialClient({self.framer} baud[{self.params.baudrate}])"

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"framer={self.framer}, timeout={self.params.timeout}>"
        )



"""Test client sync diag."""
from itertools import count
import socket

import unittest
from unittest.mock import MagicMock, patch
from test.test_client_sync import mockSocket

from pymodbus.client.sync_diag import ModbusTcpDiagClient, get_client
from pymodbus.exceptions import ConnectionException


# ---------------------------------------------------------------------------#
# Fixture
# ---------------------------------------------------------------------------#


class SynchronousDiagnosticClientTest(unittest.TestCase):
    """Unittest for the pymodbus.client.sync_diag module.

    It is a copy of parts of the test for the TCP class in the pymodbus.client
    module, as it should operate identically and only log some additional
    lines.
    """

    # -----------------------------------------------------------------------#
    # Test TCP Diagnostic Client
    # -----------------------------------------------------------------------#

    def test_syn_tcp_diag_client_instantiation(self):
        """Test sync tcp diag client."""
        client = get_client()
        self.assertNotEqual(client, None)

    def test_basic_syn_tcp_diag_client(self):
        """Test the basic methods for the tcp sync diag client"""
        # connect/disconnect
        client = ModbusTcpDiagClient()
        client.socket = mockSocket()
        self.assertTrue(client.connect())
        client.close()

    def test_tcp_diag_client_connect(self):
        """Test the tcp sync diag client connection method"""
        with patch.object(socket, "create_connection") as mock_method:
            mock_method.return_value = object()
            client = ModbusTcpDiagClient()
            self.assertTrue(client.connect())

        with patch.object(socket, "create_connection") as mock_method:
            mock_method.side_effect = socket.error()
            client = ModbusTcpDiagClient()
            self.assertFalse(client.connect())

    @patch("pymodbus.client.tcp.time")
    @patch("pymodbus.client.sync_diag.time")
    @patch("pymodbus.client.tcp.select")
    def test_tcp_diag_client_recv(self, mock_select, mock_diag_time, mock_time):
        """Test the tcp sync diag client receive method"""
        mock_select.select.return_value = [True]
        mock_time.time.side_effect = count()
        mock_diag_time.time.side_effect = count()
        client = ModbusTcpDiagClient()
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )

        client.socket = mockSocket()
        # Test logging of non-delayed responses
        client.socket.mock_store(b"\x00")
        self.assertIn(b"\x00", client.recv(None))
        client.socket = mockSocket()
        client.socket.mock_store(b"\x00")
        self.assertEqual(b"\x00", client.recv(1))

        # Fool diagnostic logger into thinking we"re running late,
        # test logging of delayed responses
        mock_diag_time.time.side_effect = count(step=3)
        client.socket.mock_store(b"\x00" * 4)
        self.assertEqual(b"\x00" * 4, client.recv(4))
        self.assertEqual(b"", client.recv(0))

        client.socket.mock_store(b"\x00\x01\x02")
        client.timeout = 3
        self.assertEqual(b"\x00\x01\x02", client.recv(3))
        client.socket.mock_store(b"\x00\x01\x02")
        self.assertEqual(b"\x00\x01", client.recv(2))
        mock_select.select.return_value = [False]
        self.assertEqual(b"", client.recv(2))
        client.socket = mockSocket()
        client.socket.mock_store(b"\x00")
        mock_select.select.return_value = [True]
        self.assertIn(b"\x00", client.recv(None))

        mock_socket = MagicMock()
        client.socket = mock_socket
        mock_socket.recv.return_value = b""
        self.assertRaises(
            ConnectionException,
            lambda: client.recv(1024),
        )
        client.socket = mockSocket()
        client.socket.mock_store(b"\x00\x01\x02")
        self.assertEqual(b"\x00\x01\x02", client.recv(1024))

    def test_tcp_diag_client_repr(self):
        """Test tcp diag client."""
        client = ModbusTcpDiagClient()
        rep = (
            f"<{client.__class__.__name__} at {hex(id(client))} "
            f"socket={client.socket}, ipaddr={client.params.host}, "
            f"port={client.params.port}, timeout={client.params.timeout}>"
        )
        self.assertEqual(repr(client), rep)



#!/usr/bin/env python3
"""This is a simple scraper.

that can be pointed at a modbus device to pull down all its values and store
them as a collection of sequential data blocks.
"""
import logging
from optparse import OptionParser
import pickle  # nosec

from twisted.internet import reactor, serialport  # pylint: disable=import-error
from twisted.internet.protocol import (
    ClientFactory,  # pylint: disable=import-error
)

from pymodbus.client.asynchronous.twisted import ModbusClientProtocol
from pymodbus.datastore import ModbusSequentialDataBlock, ModbusSlaveContext
from pymodbus.factory import ClientDecoder

# --------------------------------------------------------------------------- #
# Choose the framer you want to use
# --------------------------------------------------------------------------- #
# from pymodbus.transaction import ModbusBinaryFramer
# from pymodbus.transaction import ModbusAsciiFramer
# from pymodbus.transaction import ModbusRtuFramer
from pymodbus.transaction import ModbusSocketFramer


# -------------------------------------------------------------------------- #
# Configure the client logging
# -------------------------------------------------------------------------- #
log = logging.getLogger("pymodbus")


# --------------------------------------------------------------------------- #
# Define some constants
# --------------------------------------------------------------------------- #
COUNT = 8  # The number of bits/registers to read at once
DELAY = 0  # The delay between subsequent reads
SLAVE = 0x01  # The slave unit id to read from

# --------------------------------------------------------------------------- #
# A simple scraper protocol
# --------------------------------------------------------------------------- #
# I tried to spread the load across the device, but feel free to modify the
# logic to suit your own purpose.
# --------------------------------------------------------------------------- #


class ScraperProtocol(ModbusClientProtocol):
    """Scraper protocol."""

    address = None

    def __init__(self, framer, endpoint):
        """Initialize our custom protocol

        :param framer: The decoder to use to process messages
        :param endpoint: The endpoint to send results to
        """
        ModbusClientProtocol.__init__(self, framer)
        self.endpoint = endpoint

    def connection_made(self):
        """Call when the client has connected to the remote server."""
        super().connectionMade()
        log.debug("Beginning the processing loop")
        self.address = self.factory.starting
        reactor.callLater(DELAY, self.scrape_holding_registers)

    def connection_lost(self, reason):
        """Call when the client disconnects from the server.

        :param reason: The reason for the disconnection
        """
        reactor.callLater(DELAY, reactor.stop)

    def scrape_holding_registers(self):
        """Defer fetching holding registers"""
        txt = f"reading holding registers: {self.address}"
        log.debug(txt)
        data = self.read_holding_registers(self.address, count=COUNT, unit=SLAVE)
        data.addCallbacks(self.scrape_discrete_inputs, self.error_handler)

    def scrape_discrete_inputs(self, response):
        """Defer fetching holding registers"""
        txt = f"reading discrete inputs: {self.address}"
        log.debug(txt)
        self.endpoint.write((3, self.address, response.registers))
        data = self.read_discrete_inputs(self.address, count=COUNT, unit=SLAVE)
        data.addCallbacks(self.scrape_input_registers, self.error_handler)

    def scrape_input_registers(self, response):
        """Defer fetching holding registers"""
        txt = f"reading discrete inputs: {self.address}"
        log.debug(txt)
        self.endpoint.write((2, self.address, response.bits))
        data = self.read_input_registers(self.address, count=COUNT, unit=SLAVE)
        data.addCallbacks(self.scrape_coils, self.error_handler)

    def scrape_coils(self, response):
        """Write values of holding registers, defer fetching coils

        :param response: The response to process
        """
        txt = f"reading coils: {self.address}"
        log.debug(txt)
        self.endpoint.write((4, self.address, response.registers))
        data = self.read_coils(self.address, count=COUNT, unit=SLAVE)
        data.addCallbacks(self.start_next_cycle, self.error_handler)

    def start_next_cycle(self, response):
        """Write values of coils, trigger next cycle

        :param response: The response to process
        """
        txt = f"starting next round: {self.address}"
        log.debug(txt)
        self.endpoint.write((1, self.address, response.bits))
        self.address += COUNT
        if self.address >= self.factory.ending:
            self.endpoint.finalize()
            self.transport.loseConnection()
        else:
            reactor.callLater(DELAY, self.scrape_holding_registers)

    def error_handler(self, failure):
        """Handle any twisted errors

        :param failure: The error to handle
        """
        log.error(failure)


# --------------------------------------------------------------------------- #
# a factory for the example protocol
# --------------------------------------------------------------------------- #
# This is used to build client protocol"s if you tie into twisted"s method
# of processing. It basically produces client instances of the underlying
# protocol::
#
#     Factory(Protocol) -> ProtocolInstance
#
# It also persists data between client instances (think protocol singleton).
# --------------------------------------------------------------------------- #
class ScraperFactory(ClientFactory):  # pylint: disable=too-few-public-methods
    """Scraper factory."""

    protocol = ScraperProtocol

    def __init__(self, framer, endpoint, query):
        """Remember things necessary for building a protocols"""
        self.framer = framer
        self.endpoint = endpoint
        self.starting, self.ending = query

    def buildProtocol(self, _):  # pylint: disable=invalid-name
        """Create a protocol and start the reading cycle"""
        protocol = self.protocol(self.framer, self.endpoint)
        protocol.factory = self  # pylint: disable=attribute-defined-outside-init
        return protocol


# --------------------------------------------------------------------------- #
# a custom client for our device
# --------------------------------------------------------------------------- #
# Twisted provides a number of helper methods for creating and starting
# clients:
# - protocol.ClientCreator
# - reactor.connectTCP
#
# How you start your client is really up to you.
# --------------------------------------------------------------------------- #
class SerialModbusClient(serialport.SerialPort):
    """Serial modbus client."""

    def __init__(self, factory, *args, **kwargs):
        """Do setup the client and start listening on the serial port

        :param factory: The factory to build clients with
        """
        protocol = factory.buildProtocol(None)
        self.decoder = ClientDecoder()
        serialport.SerialPort.__init__(self, protocol, *args, **kwargs)


# --------------------------------------------------------------------------- #
# a custom endpoint for our results
# --------------------------------------------------------------------------- #
# An example line reader, this can replace with:
# - the TCP protocol
# - a context recorder
# - a database or file recorder
# --------------------------------------------------------------------------- #
class LoggingContextReader:
    """Logging context reader."""

    def __init__(self, output):
        """Initialize a new instance of the logger

        :param output: The output file to save to
        """
        self.output = output
        self.context = ModbusSlaveContext(
            di=ModbusSequentialDataBlock.create(),
            co=ModbusSequentialDataBlock.create(),
            hr=ModbusSequentialDataBlock.create(),
            ir=ModbusSequentialDataBlock.create(),
        )

    def write(self, response):
        """Handle the next modbus response

        :param response: The response to process
        """
        txt = f"Read Data: {str(response)}"
        log.info(txt)
        file_ptr, address, values = response
        self.context.setValues(file_ptr, address, values)

    def finalize(self):
        """Finalize."""
        with open(self.output, "w") as handle:
            pickle.dump(self.context, handle)


# -------------------------------------------------------------------------- #
# Main start point
# -------------------------------------------------------------------------- #
def get_options():
    """Parse the command line options

    :returns: The options manager
    """
    parser = OptionParser()

    parser.add_option(
        "-o",
        "--output",
        help="The resulting output file for the scrape",
        dest="output",
        default="datastore.pickle",
    )

    parser.add_option(
        "-p",
        "--port",
        help="The port to connect to",
        type="int",
        dest="port",
        default=502,
    )

    parser.add_option(
        "-s", "--server", help="The server to scrape", dest="host", default="127.0.0.1"
    )

    parser.add_option(
        "-r",
        "--range",
        help="The address range to scan",
        dest="query",
        default="0:1000",
    )

    parser.add_option(
        "-d",
        "--debug",
        help="Enable debug tracing",
        action="store_true",
        dest="debug",
        default=False,
    )

    (opt, _) = parser.parse_args()
    return opt


def main():
    """Run main runner function"""
    options = get_options()

    if options.debug:
        try:
            log.setLevel(logging.DEBUG)
        except Exception:
            print("Logging is not supported on this system")

    # split the query into a starting and ending range
    query = [int(p) for p in options.query.split(":")]

    try:
        log.debug("Initializing the client")
        framer = ModbusSocketFramer(ClientDecoder())
        reader = LoggingContextReader(options.output)
        factory = ScraperFactory(framer, reader, query)

        # how to connect based on TCP vs Serial clients
        if isinstance(framer, ModbusSocketFramer):
            reactor.connectTCP(options.host, options.port, factory)
        else:
            SerialModbusClient(factory, options.port, reactor)

        log.debug("Starting the client")
        reactor.run()
        log.debug("Finished scraping the client")
    except Exception as exc:
        print(exc)


# --------------------------------------------------------------------------- #
# Main jumper
# --------------------------------------------------------------------------- #


if __name__ == "__main__":
    main()



"""Modbus Request/Response Decoder Factories.

The following factories make it easy to decode request/response messages.
To add a new request/response pair to be decodeable by the library, simply
add them to the respective function lookup table (order doesn"t matter, but
it does help keep things organized).

Regardless of how many functions are added to the lookup, O(1) behavior is
kept as a result of a pre-computed lookup dictionary.
"""
# pylint: disable=missing-type-doc
import logging

from pymodbus.bit_read_message import (
    ReadCoilsRequest,
    ReadCoilsResponse,
    ReadDiscreteInputsRequest,
    ReadDiscreteInputsResponse,
)
from pymodbus.bit_write_message import (
    WriteMultipleCoilsRequest,
    WriteMultipleCoilsResponse,
    WriteSingleCoilRequest,
    WriteSingleCoilResponse,
)
from pymodbus.diag_message import (
    ChangeAsciiInputDelimiterRequest,
    ChangeAsciiInputDelimiterResponse,
    ClearCountersRequest,
    ClearCountersResponse,
    ClearOverrunCountRequest,
    ClearOverrunCountResponse,
    DiagnosticStatusRequest,
    DiagnosticStatusResponse,
    ForceListenOnlyModeRequest,
    ForceListenOnlyModeResponse,
    GetClearModbusPlusRequest,
    GetClearModbusPlusResponse,
    RestartCommunicationsOptionRequest,
    RestartCommunicationsOptionResponse,
    ReturnBusCommunicationErrorCountRequest,
    ReturnBusCommunicationErrorCountResponse,
    ReturnBusExceptionErrorCountRequest,
    ReturnBusExceptionErrorCountResponse,
    ReturnBusMessageCountRequest,
    ReturnBusMessageCountResponse,
    ReturnDiagnosticRegisterRequest,
    ReturnDiagnosticRegisterResponse,
    ReturnIopOverrunCountRequest,
    ReturnIopOverrunCountResponse,
    ReturnQueryDataRequest,
    ReturnQueryDataResponse,
    ReturnSlaveBusCharacterOverrunCountRequest,
    ReturnSlaveBusCharacterOverrunCountResponse,
    ReturnSlaveBusyCountRequest,
    ReturnSlaveBusyCountResponse,
    ReturnSlaveMessageCountRequest,
    ReturnSlaveMessageCountResponse,
    ReturnSlaveNAKCountRequest,
    ReturnSlaveNAKCountResponse,
    ReturnSlaveNoReponseCountResponse,
    ReturnSlaveNoResponseCountRequest,
)
from pymodbus.exceptions import MessageRegisterException, ModbusException
from pymodbus.file_message import (
    ReadFifoQueueRequest,
    ReadFifoQueueResponse,
    ReadFileRecordRequest,
    ReadFileRecordResponse,
    WriteFileRecordRequest,
    WriteFileRecordResponse,
)
from pymodbus.interfaces import IModbusDecoder
from pymodbus.mei_message import (
    ReadDeviceInformationRequest,
    ReadDeviceInformationResponse,
)
from pymodbus.other_message import (
    GetCommEventCounterRequest,
    GetCommEventCounterResponse,
    GetCommEventLogRequest,
    GetCommEventLogResponse,
    ReadExceptionStatusRequest,
    ReadExceptionStatusResponse,
    ReportSlaveIdRequest,
    ReportSlaveIdResponse,
)
from pymodbus.pdu import (
    ExceptionResponse,
    IllegalFunctionRequest,
    ModbusExceptions as ecode,
    ModbusRequest,
    ModbusResponse,
)
from pymodbus.register_read_message import (
    ReadHoldingRegistersRequest,
    ReadHoldingRegistersResponse,
    ReadInputRegistersRequest,
    ReadInputRegistersResponse,
    ReadWriteMultipleRegistersRequest,
    ReadWriteMultipleRegistersResponse,
)
from pymodbus.register_write_message import (
    MaskWriteRegisterRequest,
    MaskWriteRegisterResponse,
    WriteMultipleRegistersRequest,
    WriteMultipleRegistersResponse,
    WriteSingleRegisterRequest,
    WriteSingleRegisterResponse,
)


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------- #
# Server Decoder
# --------------------------------------------------------------------------- #
class ServerDecoder(IModbusDecoder):
    """Request Message Factory (Server).

    To add more implemented functions, simply add them to the list
    """

    __function_table = [
        ReadHoldingRegistersRequest,
        ReadDiscreteInputsRequest,
        ReadInputRegistersRequest,
        ReadCoilsRequest,
        WriteMultipleCoilsRequest,
        WriteMultipleRegistersRequest,
        WriteSingleRegisterRequest,
        WriteSingleCoilRequest,
        ReadWriteMultipleRegistersRequest,
        DiagnosticStatusRequest,
        ReadExceptionStatusRequest,
        GetCommEventCounterRequest,
        GetCommEventLogRequest,
        ReportSlaveIdRequest,
        ReadFileRecordRequest,
        WriteFileRecordRequest,
        MaskWriteRegisterRequest,
        ReadFifoQueueRequest,
        ReadDeviceInformationRequest,
    ]
    __sub_function_table = [
        ReturnQueryDataRequest,
        RestartCommunicationsOptionRequest,
        ReturnDiagnosticRegisterRequest,
        ChangeAsciiInputDelimiterRequest,
        ForceListenOnlyModeRequest,
        ClearCountersRequest,
        ReturnBusMessageCountRequest,
        ReturnBusCommunicationErrorCountRequest,
        ReturnBusExceptionErrorCountRequest,
        ReturnSlaveMessageCountRequest,
        ReturnSlaveNoResponseCountRequest,
        ReturnSlaveNAKCountRequest,
        ReturnSlaveBusyCountRequest,
        ReturnSlaveBusCharacterOverrunCountRequest,
        ReturnIopOverrunCountRequest,
        ClearOverrunCountRequest,
        GetClearModbusPlusRequest,
        ReadDeviceInformationRequest,
    ]

    def __init__(self):
        """Initialize the client lookup tables."""
        functions = {f.function_code for f in self.__function_table}
        self.__lookup = {f.function_code: f for f in self.__function_table}
        self.__sub_lookup = {f: {} for f in functions}
        for f in self.__sub_function_table:
            self.__sub_lookup[f.function_code][f.sub_function_code] = f

    def decode(self, message):
        """Decode a request packet

        :param message: The raw modbus request packet
        :return: The decoded modbus message or None if error
        """
        try:
            return self._helper(message)
        except ModbusException as exc:
            txt = f"Unable to decode request {exc}"
            _logger.warning(txt)
        return None

    def lookupPduClass(self, function_code):
        """Use `function_code` to determine the class of the PDU.

        :param function_code: The function code specified in a frame.
        :returns: The class of the PDU that has a matching `function_code`.
        """
        return self.__lookup.get(function_code, ExceptionResponse)

    def _helper(self, data):
        """Generate the correct request object from a valid request packet.

        This decodes from a list of the currently implemented request types.

        :param data: The request packet to decode
        :returns: The decoded request or illegal function request object
        """
        function_code = int(data[0])
        if not (request := self.__lookup.get(function_code, lambda: None)()):
            txt = f"Factory Request[{function_code}]"
            _logger.debug(txt)
            request = IllegalFunctionRequest(function_code)
        else:
            fc_string = "%s: %s" % (  # pylint: disable=consider-using-f-string
                str(self.__lookup[function_code])  # pylint: disable=use-maxsplit-arg
                .split(".")[-1]
                .rstrip('">"'),
                function_code,
            )
            txt = f"Factory Request[{fc_string}]"
            _logger.debug(txt)
        request.decode(data[1:])

        if hasattr(request, "sub_function_code"):
            lookup = self.__sub_lookup.get(request.function_code, {})
            if subtype := lookup.get(request.sub_function_code, None):
                request.__class__ = subtype

        return request

    def register(self, function=None):
        """Register a function and sub function class with the decoder.

        :param function: Custom function class to register
        :raises MessageRegisterException:
        """
        if function and not issubclass(function, ModbusRequest):
            raise MessageRegisterException(
                f'"{function.__class__.__name__}" is Not a valid Modbus Message'
                ". Class needs to be derived from "
                "`pymodbus.pdu.ModbusRequest` "
            )
        self.__lookup[function.function_code] = function
        if hasattr(function, "sub_function_code"):
            if function.function_code not in self.__sub_lookup:
                self.__sub_lookup[function.function_code] = {}
            self.__sub_lookup[function.function_code][
                function.sub_function_code
            ] = function


# --------------------------------------------------------------------------- #
# Client Decoder
# --------------------------------------------------------------------------- #
class ClientDecoder(IModbusDecoder):
    """Response Message Factory (Client).

    To add more implemented functions, simply add them to the list
    """

    __function_table = [
        ReadHoldingRegistersResponse,
        ReadDiscreteInputsResponse,
        ReadInputRegistersResponse,
        ReadCoilsResponse,
        WriteMultipleCoilsResponse,
        WriteMultipleRegistersResponse,
        WriteSingleRegisterResponse,
        WriteSingleCoilResponse,
        ReadWriteMultipleRegistersResponse,
        DiagnosticStatusResponse,
        ReadExceptionStatusResponse,
        GetCommEventCounterResponse,
        GetCommEventLogResponse,
        ReportSlaveIdResponse,
        ReadFileRecordResponse,
        WriteFileRecordResponse,
        MaskWriteRegisterResponse,
        ReadFifoQueueResponse,
        ReadDeviceInformationResponse,
    ]
    __sub_function_table = [
        ReturnQueryDataResponse,
        RestartCommunicationsOptionResponse,
        ReturnDiagnosticRegisterResponse,
        ChangeAsciiInputDelimiterResponse,
        ForceListenOnlyModeResponse,
        ClearCountersResponse,
        ReturnBusMessageCountResponse,
        ReturnBusCommunicationErrorCountResponse,
        ReturnBusExceptionErrorCountResponse,
        ReturnSlaveMessageCountResponse,
        ReturnSlaveNoReponseCountResponse,
        ReturnSlaveNAKCountResponse,
        ReturnSlaveBusyCountResponse,
        ReturnSlaveBusCharacterOverrunCountResponse,
        ReturnIopOverrunCountResponse,
        ClearOverrunCountResponse,
        GetClearModbusPlusResponse,
        ReadDeviceInformationResponse,
    ]

    def __init__(self):
        """Initialize the client lookup tables."""
        functions = {f.function_code for f in self.__function_table}
        self.__lookup = {f.function_code: f for f in self.__function_table}
        self.__sub_lookup = {f: {} for f in functions}
        for f in self.__sub_function_table:
            self.__sub_lookup[f.function_code][f.sub_function_code] = f

    def lookupPduClass(self, function_code):
        """Use `function_code` to determine the class of the PDU.

        :param function_code: The function code specified in a frame.
        :returns: The class of the PDU that has a matching `function_code`.
        """
        return self.__lookup.get(function_code, ExceptionResponse)

    def decode(self, message):
        """Decode a response packet.

        :param message: The raw packet to decode
        :return: The decoded modbus message or None if error
        """
        try:
            return self._helper(message)
        except ModbusException as exc:
            txt = f"Unable to decode response {exc}"
            _logger.error(txt)

        except Exception as exc:  # pylint: disable=broad-except
            _logger.error(exc)
        return None

    def _helper(self, data):
        """Generate the correct response object from a valid response packet.

        This decodes from a list of the currently implemented request types.

        :param data: The response packet to decode
        :returns: The decoded request or an exception response object
        :raises ModbusException:
        """
        fc_string = function_code = int(data[0])
        if function_code in self.__lookup:
            fc_string = "%s: %s" % (  # pylint: disable=consider-using-f-string
                str(self.__lookup[function_code])  # pylint: disable=use-maxsplit-arg
                .split(".")[-1]
                .rstrip('">"'),
                function_code,
            )
        txt = f"Factory Response[{fc_string}]"
        _logger.debug(txt)
        response = self.__lookup.get(function_code, lambda: None)()
        if function_code > 0x80:
            code = function_code & 0x7F  # strip error portion
            response = ExceptionResponse(code, ecode.IllegalFunction)
        if not response:
            raise ModbusException(f"Unknown response {function_code}")
        response.decode(data[1:])

        if hasattr(response, "sub_function_code"):
            lookup = self.__sub_lookup.get(response.function_code, {})
            if subtype := lookup.get(response.sub_function_code, None):
                response.__class__ = subtype

        return response

    def register(self, function):
        """Register a function and sub function class with the decoder."""
        if function and not issubclass(function, ModbusResponse):
            raise MessageRegisterException(
                f'"{function.__class__.__name__}" is Not a valid Modbus Message'
                ". Class needs to be derived from "
                "`pymodbus.pdu.ModbusResponse` "
            )
        self.__lookup[function.function_code] = function
        if hasattr(function, "sub_function_code"):
            if function.function_code not in self.__sub_lookup:
                self.__sub_lookup[function.function_code] = {}
            self.__sub_lookup[function.function_code][
                function.sub_function_code
            ] = function


# --------------------------------------------------------------------------- #
# Exported symbols
# --------------------------------------------------------------------------- #


__all__ = ["ServerDecoder", "ClientDecoder"]



"""Test server context."""
import unittest

from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext
from pymodbus.exceptions import NoSuchSlaveException


class ModbusServerSingleContextTest(unittest.TestCase):
    """This is the unittest for the pymodbus.datastore.ModbusServerContext using a single slave context."""

    def setUp(self):
        """Set up the test environment"""
        self.slave = ModbusSlaveContext()
        self.context = ModbusServerContext(slaves=self.slave, single=True)

    def tearDown(self):
        """Clean up the test environment"""
        del self.context

    def test_single_context_gets(self):
        """Test getting on a single context"""
        for slave_id in range(0, 0xFF):
            self.assertEqual(self.slave, self.context[slave_id])

    def test_single_context_deletes(self):
        """Test removing on multiple context"""

        def _test():
            del self.context[0x00]

        self.assertRaises(NoSuchSlaveException, _test)

    def test_single_context_iter(self):
        """Test iterating over a single context"""
        expected = (0, self.slave)
        for slave in self.context:
            self.assertEqual(slave, expected)

    def test_single_context_default(self):
        """Test that the single context default values work"""
        self.context = ModbusServerContext()
        slave = self.context[0x00]
        self.assertEqual(slave, {})

    def test_single_context_set(self):
        """Test a setting a single slave context"""
        slave = ModbusSlaveContext()
        self.context[0x00] = slave
        actual = self.context[0x00]
        self.assertEqual(slave, actual)

    def test_single_context_register(self):
        """Test single context register."""
        request_db = [1, 2, 3]
        slave = ModbusSlaveContext()
        slave.register(0xFF, "custom_request", request_db)
        self.assertEqual(slave.store["custom_request"], request_db)
        self.assertEqual(slave.decode(0xFF), "custom_request")


class ModbusServerMultipleContextTest(unittest.TestCase):
    """This is the unittest for the pymodbus.datastore.ModbusServerContext using multiple slave contexts."""

    def setUp(self):
        """Set up the test environment"""
        self.slaves = {id: ModbusSlaveContext() for id in range(10)}
        self.context = ModbusServerContext(slaves=self.slaves, single=False)

    def tearDown(self):
        """Clean up the test environment"""
        del self.context

    def test_multiple_context_gets(self):
        """Test getting on multiple context"""
        for slave_id in range(0, 10):
            self.assertEqual(self.slaves[slave_id], self.context[slave_id])

    def test_multiple_context_deletes(self):
        """Test removing on multiple context"""
        del self.context[0x00]
        self.assertRaises(NoSuchSlaveException, lambda: self.context[0x00])

    def test_multiple_context_iter(self):
        """Test iterating over multiple context"""
        for slave_id, slave in self.context:
            self.assertEqual(slave, self.slaves[slave_id])
            self.assertTrue(slave_id in self.context)

    def test_multiple_context_default(self):
        """Test that the multiple context default values work"""
        self.context = ModbusServerContext(single=False)
        self.assertRaises(NoSuchSlaveException, lambda: self.context[0x00])

    def test_multiple_context_set(self):
        """Test a setting multiple slave contexts"""
        slaves = {id: ModbusSlaveContext() for id in range(10)}
        for slave_id, slave in iter(slaves.items()):
            self.context[slave_id] = slave
        for slave_id, slave in iter(slaves.items()):
            actual = self.context[slave_id]
            self.assertEqual(slave, actual)



"""Define Datastore."""
from pymodbus.datastore.database.redis_datastore import RedisSlaveContext
from pymodbus.datastore.database.sql_datastore import SqlSlaveContext


# ---------------------------------------------------------------------------#
#  Exported symbols
# ---------------------------------------------------------------------------#
__all__ = ["SqlSlaveContext", "RedisSlaveContext"]



"""Define datastore."""
from pymodbus.datastore.context import ModbusServerContext, ModbusSlaveContext
from pymodbus.datastore.store import (
    ModbusSequentialDataBlock,
    ModbusSparseDataBlock,
)


# ---------------------------------------------------------------------------#
#  Exported symbols
# ---------------------------------------------------------------------------#
__all__ = [
    "ModbusSequentialDataBlock",
    "ModbusSparseDataBlock",
    "ModbusSlaveContext",
    "ModbusServerContext",
]
